import base64, binascii, struct, time, fnmatch, ipaddress

class DNSError(Exception):
    pass

class BufferError(Exception):
    pass

class DNSLabel(object):
    def __init__(self, label, origin=None):
        if isinstance(label, DNSLabel):
            self.label = label.label
        elif isinstance(label, (list, tuple)):
            self.label = tuple(label)
        else:
            final = False
            if not label or label in (b'.', '.'):
                self.label = ()
                final = bool(label)
            elif isinstance(label, str):
                self.label = tuple(label.encode("idna").rstrip(b".").split(b"."))
                final = label.endswith(".")
            else:
                self.label = tuple(label.rstrip(b".").split(b"."))
                final = label.endswith(b".")
            if not final and origin is not None:
                self.label += DNSLabel(origin).label
    def matchGlob(self, pattern):
        return fnmatch.fnmatch(str(self).lower(), str(DNSLabel(pattern)).lower())
    def __str__(self):
        return ".".join(s.decode() for s in self.label)+"."
    def __repr__(self):
        return "<DNSLabel: '%s'>" % str(self)
    def __hash__(self):
        return hash(self.label)
    def __len__(self):
        return len(b'.'.join(self.label))

class Buffer(object):
    def __init__(self, data=b''):
        self.data = bytearray(data)
        self.offset = 0
        self.names = {}
    def remaining(self):
        return len(self.data) - self.offset
    def get(self, length):
        if length > self.remaining():
            raise BufferError("Not enough bytes [offset=%d,remaining=%d,requested=%d]" %
                    (self.offset, self.remaining(), length))
        start = self.offset
        self.offset += length
        return bytes(self.data[start:self.offset])
    def get1(self):
        if self.remaining() == 0:
            raise BufferError("Not enough bytes [offset=%d,remaining=%d,requested=%d]" %
                    (self.offset, self.remaining(), 1))
        self.offset += 1
        return self.data[self.offset-1]
    def pack(self, fmt, *args):
        for i, d in zip(fmt, args):
            if i == 'B':
                self.append(struct.pack('!B', d))
            elif i == 'H':
                self.append(struct.pack('!H', d))
            elif i == 'I':
                self.append(struct.pack('!I', d))
            elif i == 'L':
                self.encode_name(d)
            elif i == 'l':
                self.encode_name_nocompress(d)
            elif i == 'S':
                self.append(struct.pack('!H', len(d)))
                self.append(d)
            elif i == 's':
                if len(d) > 255:
                    raise BufferError("Error packing struct '%s' <%s>: too long string %d" % (fmt, data, len(d)))
                self.append(struct.pack('!B', len(d)))
                self.append(d)
            elif i in ('4', '6'):
                self.append(d.packed)
            elif i == 'D':
                self.append(d)
            else:
                raise BufferError("Error packing struct '%s' <%s>: unknown fmt '%s'" % (fmt, data, i))
    def append(self, s):
        self.offset += len(s)
        self.data.extend(s)
    def update(self, ptr, fmt, *args):
        s = Buffer()
        s.pack(fmt, *args)
        self.data[ptr:ptr+len(s)] = s.data
    def unpack(self, fmt):
        data = []
        for i in fmt:
            if i == 'B':
                data.append(self.get1())
            elif i == 'H':
                data.extend(struct.unpack('!H', self.get(2)))
            elif i == 'I':
                data.extend(struct.unpack('!I', self.get(4)))
            elif i in ('L', 'l'):
                data.append(self.decode_name())
            elif i == 'S':
                length, = struct.unpack('!H', self.get(2))
                data.append(self.get(length))
            elif i == 's':
                length = self.get1()
                data.append(self.get(length))
            elif i == '4':
                data.append(ipaddress.IPv4Address(self.get(4)))
            elif i == '6':
                data.append(ipaddress.IPv6Address(self.get(16)))
            elif i == 'D':
                length = self.remaining()
                data.append(self.get(length))
            else:
                raise BufferError("Error unpacking struct '%s' <%s>" % (fmt, binascii.hexlify(self.data).decode()))
        return data
    def __len__(self):
        return len(self.data)
    def decode_name(self):
        label = []
        offsets = [self.offset]
        length = self.get1()
        while 0 < length < 0xC0:
            label.append(self.get(length))
            offsets.append(self.offset)
            length = self.get1()
        if length:
            pointer = ((length&0x3F)<<8) | self.get1()
            if pointer in self.names:
                label.extend(self.names[pointer])
            else:
                raise BufferError("Invalid pointer in DNSLabel [offset=%d,pointer=%d,length=%d]" %
                        (self.offset, pointer, len(self.data)))
        for i, j in enumerate(offsets):
            self.names[j] = label[i:]
        return DNSLabel(label)
    def encode_name(self, name):
        name = list(DNSLabel(name).label)
        while name and tuple(name) not in self.names:
            self.names[tuple(name)] = self.offset
            self.pack("s", name.pop(0))
        if name:
            self.pack("H", self.names[tuple(name)]|0xC000)
        else:
            self.append(b'\x00')
    def encode_name_nocompress(self, name):
        for element in DNSLabel(name).label:
            self.pack("s", element)
        self.append(b'\x00')

# DNS codes
QTYPE =  {1:'A', 2:'NS', 5:'CNAME', 6:'SOA', 12:'PTR', 15:'MX', 16:'TXT', 17:'RP', 18:'AFSDB',
          24:'SIG', 25:'KEY', 28:'AAAA', 29:'LOC', 33:'SRV', 35:'NAPTR', 36:'KX', 37:'CERT',
          39:'DNAME', 41:'OPT', 42:'APL', 43:'DS', 44:'SSHFP', 45:'IPSECKEY', 46:'RRSIG',
          47:'NSEC', 48:'DNSKEY', 49:'DHCID', 50:'NSEC3', 51:'NSEC3PARAM', 52:'TLSA', 55:'HIP',
          99:'SPF', 249:'TKEY', 250:'TSIG', 251:'IXFR', 252:'AXFR', 255:'ANY', 257:'TYPE257',
          32768:'TA', 32769:'DLV'}
QTYPE_OPT = 41
QTYPE_R = {j:i for i,j in QTYPE.items()}

class DNSRecord(object):
    @classmethod
    def unpack(cls, packet):
        try:
            buffer = Buffer(packet)
            header = DNSHeader.unpack(buffer)
            questions = []
            rr = []
            auth = []
            ar = []
            for i in range(header.q):
                questions.append(DNSQuestion.unpack(buffer))
            for i in range(header.a):
                rr.append(RR.unpack(buffer))
            for i in range(header.auth):
                auth.append(RR.unpack(buffer))
            for i in range(header.ar):
                ar.append(RR.unpack(buffer))
            return cls(header, questions, rr, auth=auth, ar=ar)
        except DNSError:
            raise
        except BufferError as e:
            raise DNSError("Error unpacking DNSRecord [offset=%d]: %s" % (buffer.offset, e))
    def __init__(self, header=None, questions=None, rr=None, q=None, a=None, auth=None, ar=None):
        self.header = header or DNSHeader()
        self.questions = questions or []
        self.rr = rr or []
        self.auth = auth or []
        self.ar = ar or []
        if q: self.questions.append(q)
        if a: self.rr.append(a)
        self.set_header_qa()
    def reply(self, ra=1, aa=1):
        header = DNSHeader(self.header.id, self.header.bitmap, 0, 0, 0, 0, qr=1, ra=ra, aa=aa)
        return DNSRecord(header, q=self.q)
    def add_answer(self, *rr):
        self.rr.extend(rr)
        self.set_header_qa()
    def set_header_qa(self):
        self.header.q = len(self.questions)
        self.header.a = len(self.rr)
        self.header.auth = len(self.auth)
        self.header.ar = len(self.ar)
    @property
    def q(self):
        return self.questions[0] if self.questions else DNSQuestion()
    @property
    def a(self):
        return self.rr[0] if self.rr else RR()
    def pack(self):
        self.set_header_qa()
        buffer = Buffer()
        self.header.pack(buffer)
        for q in self.questions:
            q.pack(buffer)
        for rr in self.rr:
            rr.pack(buffer)
        for auth in self.auth:
            auth.pack(buffer)
        for ar in self.ar:
            ar.pack(buffer)
        return buffer.data

make_time = lambda d: int(time.mktime(time.strptime(d+'GMT',"%Y%m%d%H%M%S%Z")))
make_hex = lambda d: binascii.unhexlify(d.encode('ascii'))
class Struct(object):
    STRUCT = ""
    ATTRS = ()
    ZONES = {}
    @classmethod
    def unpack(cls, buffer):
        try:
            return cls(*buffer.unpack(cls.STRUCT))
        except BufferError as e:
            raise DNSError("Error unpacking %s [offset=%d]: %s" % (cls.__name__, buffer.offset, e))
    @classmethod
    def fromZone(cls, rd, origin=None):
        args = []
        idx = 0
        while idx < len(cls.ATTRS):
            i = cls.STRUCT[idx]
            t = cls.ATTRS[idx]
            d = rd[idx]
            f = cls.ZONES.get(t)
            if f:
                j = f(d)
            elif i in ('B', 'H', 'I'):
                j = int(d)
            elif i in ('L', 'l'):
                j = DNSLabel(d, origin)
            elif i in ('s', 'S'):
                j = d.encode()
            elif i == 'D':
                j = base64.b64decode(("".join(rd[idx:])).encode('ascii'))
            elif i == '4':
                j = ipaddress.IPv4Address(d.rstrip('.'))
            elif i == '6':
                j = ipaddress.IPv6Address(d)
            else:
                raise DNSError("Unknown fmt '%s' in '%s'" % (i, cls.ATTRS))
            args.append(j)
            idx += 1
        return cls(*args)
    def __init__(self, *args, **kw):
        for i, d in zip(self.ATTRS, args):
            setattr(self, i, d)
        for i, d in kw.items():
            setattr(self, i, d)
    def pack(self, buffer):
        args = [getattr(self, i) for i in self.ATTRS]
        buffer.pack(self.STRUCT, *args)
    def __str__(self):
        values = [str(getattr(self, i)) for i in self.ATTRS]
        if len(values) == 1:
            return values[0]
        else:
            return str(values)

class DNSHeader(Struct):
    STRUCT = "HHHHHH"
    ATTRS = ('id', 'bitmap', 'q', 'a', 'auth', 'ar')
    def get_qr(self):
        return (self.bitmap>>15) & 1
    def set_qr(self, val):
        self.bitmap = self.bitmap | (1<<15)
    qr = property(get_qr, set_qr)
    def get_aa(self):
        return (self.bitmap>>10) & 1
    def set_aa(self,val):
        self.bitmap = self.bitmap | (1<<10)
    aa = property(get_aa,set_aa)
    def get_ra(self):
        return (self.bitmap>>7) & 1
    def set_ra(self, val):
        self.bitmap = self.bitmap | (1<<7)
    ra = property(get_ra, set_ra)

class DNSQuestion(Struct):
    STRUCT = "LHH"
    ATTRS = ('qname', 'qtype', 'qclass')
    def __eq__(self, other):
        return str(self.qname) == str(other.qname) and self.qtype == other.qtype and self.qclass == other.qclass
    def __hash__(self):
        return hash((str(self.qname), self.qtype, self.qclass))

class EDNSOption(Struct):
    STRUCT = "HS"
    ATTRS = ('code', 'data')

class RR(object):
    CLASS = {'IN':1, 'CS':2, 'CH':3, 'Hesiod':4, 'None':254, '*':255}
    @classmethod
    def unpack(cls, buffer):
        try:
            rname, rtype, rclass, ttl, buflen = buffer.unpack("LHHIH")
            if rtype == QTYPE_OPT:
                buf = Buffer(buffer.get(buflen))
                rdata = []
                while buf.remaining() > 4:
                    rdata.append(EDNSOption.unpack(buf))
            elif buflen:
                rdata = RDMAP.get(QTYPE.get(rtype), RD).unpack(buffer)
            else:
                rdata = ''
            return cls(rname, rtype, rclass, ttl, rdata)
        except BufferError as e:
            raise DNSError("Error unpacking RR [offset=%d]: %s" % (buffer.offset, e))
    @classmethod
    def fromZone(cls, zone, origin=None, ttl=0):
        for line in zone.split('\n'):
            rr = line.strip().split()
            if not rr: continue
            label = DNSLabel(rr.pop(0), origin)
            ttl = int(rr.pop(0)) if rr[0].isdigit() else ttl
            rclass = rr.pop(0) if rr[0] in ('IN','CS','CH') else 'IN'
            rtype = rr.pop(0)
            rdata = rr
            rd = RDMAP.get(rtype, RD)
            yield RR(rname=label, ttl=ttl, rclass=cls.CLASS[rclass],
                     rtype=QTYPE_R.get(rtype), rdata=rd.fromZone(rdata, origin))
    def __init__(self, rname=None, rtype=1, rclass=1, ttl=0, rdata=None):
        self.rname = rname
        self.rtype = rtype
        self.rclass = rclass
        self.ttl = ttl
        self.rdata = rdata
        if self.rtype == QTYPE_OPT:
            self.edns_len = self.rclass
            self.edns_do = self.ttl & 0x7FFF
            self.edns_ver = (self.ttl>>16) & 0xFF
            self.edns_rcode = (self.ttl>>24) & 0xFF
    def pack(self, buffer):
        buffer.pack("LHHIH", self.rname, self.rtype, self.rclass, self.ttl, 0)
        start = buffer.offset
        if self.rtype == QTYPE_OPT:
            for opt in self.rdata:
                opt.pack(buffer)
        else:
            self.rdata.pack(buffer)
        buffer.update(start-2, "H", buffer.offset-start)

class RD(Struct):
    STRUCT = "D"
    ATTRS = ('data',)
    ZONES = {'data': make_hex}

class TXT(Struct):
    STRUCT = "s"
    ATTRS = ('data',)

class A(Struct):
    STRUCT = "4"
    ATTRS = ('data',)

class AAAA(Struct):
    STRUCT = "6"
    ATTRS = ('data',)

class MX(Struct):
    STRUCT = "HL"
    ATTRS = ('preference', 'label')

class CNAME(Struct):
    STRUCT = "L"
    ATTRS = ('label',)

class DNAME(Struct):
    STRUCT = "L"
    ATTRS = ('label',)

class PTR(CNAME):
    pass

class NS(CNAME):
    pass

class SOA(Struct):
    STRUCT = "LLIIIII"
    ATTRS = ('mname', 'rname', 't0', 't1', 't2', 't3', 't4')

class SRV(Struct):
    STRUCT = "HHHL"
    ATTRS = ('priority', 'weight', 'port', 'target')

class NAPTR(Struct):
    STRUCT = "HHsssL"
    ATTRS = ('order', 'preference', 'flags', 'service', 'regexp', 'replacement')

class DNSKEY(Struct):
    STRUCT = "HBBD"
    ATTRS = ('flags', 'protocol', 'algorithm', 'key')

class RRSIG(Struct):
    STRUCT = "HBBIIIHlD"
    ATTRS = ('covered', 'algorithm', 'labels', 'orig_ttl', 'sig_exp', 'sig_inc', 'key_tag', 'name', 'sig')
    ZONES = {'sig_exp': make_time, 'sig_inc': make_time}

RDMAP = dict(CNAME=CNAME, DNAME=DNAME, A=A, AAAA=AAAA, TXT=TXT, MX=MX, PTR=PTR, SOA=SOA, NS=NS,
             NAPTR=NAPTR, SRV=SRV, DNSKEY=DNSKEY, RRSIG=RRSIG)

class DNSCache(object):
    def __init__(self):
        self.cache = {}
        self.rdns = {}
    def query(self, record):
        if len(record.questions) == 1:
            if record.q in self.cache:
                answer = self.cache[record.q]
                answer.header.id = record.header.id
                for r in answer.rr:
                    if r.rtype in (1, 5, 39) and r.rclass == 1:
                        self.rdns[str(r.rdata)] = str(r.rname)
                return answer
    def answer(self, record):
        if len(record.questions) == 1 and record.rr:
            self.cache[record.q] = record
        for r in record.rr:
            if r.rtype in (1, 5, 39) and r.rclass == 1:
                self.rdns[str(r.rdata)] = str(r.rname)
    def ip2domain(self, ip):
        while ip in self.rdns:
            ip = self.rdns[ip]
        return ip.rstrip('.')



import argparse, asyncio, io, os, enum, struct, collections, hashlib, ipaddress, socket, random, base64, itertools, hmac, heapq
import pproxy
from . import enums, message, crypto, ip
from .__doc__ import *

class State(enum.Enum):
    INITIAL = 0
    SA_SENT = 1
    ESTABLISHED = 2
    DELETED = 3
    KE_SENT = 4
    HASH_SENT = 5
    AUTH_SET = 6
    CONF_SENT = 7
    CHILD_SA_SENT = 8

class ChildSa:
    def __init__(self, spi_in, spi_out, crypto_in, crypto_out):
        self.spi_in = spi_in
        self.spi_out = spi_out
        self.crypto_in = crypto_in
        self.crypto_out = crypto_out
        self.msgid_in = 1
        self.msgid_out = 1
        self.msgwin_in = set()
        self.child = None
    def incr_msgid_in(self):
        self.msgid_in += 1
        while self.msgid_in in self.msgwin_in:
            self.msgwin_in.discard(self.msgid_in)
            self.msgid_in += 1

class IKEv1Session:
    all_child_sa = {}
    def __init__(self, args, sessions, peer_spi):
        self.args = args
        self.sessions = sessions
        self.my_spi = os.urandom(8)
        self.peer_spi = peer_spi
        self.crypto = None
        self.my_nonce = os.urandom(32)
        self.peer_nonce = None
        self.child_sa = []
        self.state = State.INITIAL
        self.sessions[self.my_spi] = self
    def response(self, exchange, payloads, message_id=0, *, crypto=None, hashmsg=None):
        if hashmsg:
            message_id = message_id or random.randrange(1<<32)
            buf = (b'' if hashmsg is True else hashmsg) + message.Message.encode_payloads(payloads)
            hash_r = self.crypto.prf.prf(self.skeyid_a, message_id.to_bytes(4, 'big') + buf)
            payloads.insert(0, message.PayloadHASH_1(hash_r))
        response = message.Message(self.peer_spi, self.my_spi, 0x10, exchange,
                                   enums.MsgFlag.NONE, message_id, payloads)
        #print(repr(response))
        return response.to_bytes(crypto=crypto)
    def verify_hash(self, request):
        payload_hash = request.payloads.pop(0)
        assert payload_hash.type == enums.Payload.HASH_1
        hash_i = self.crypto.prf.prf(self.skeyid_a, request.message_id.to_bytes(4, 'big') + message.Message.encode_payloads(request.payloads))
        assert hash_i == payload_hash.data
    def xauth_init(self):
        attrs = { enums.CPAttrType.XAUTH_TYPE: 0,
                  enums.CPAttrType.XAUTH_USER_NAME: b'',
                  enums.CPAttrType.XAUTH_USER_PASSWORD: b'',
                }
        response_payloads = [message.PayloadCP_1(enums.CFGType.CFG_REQUEST, attrs)]
        return self.response(enums.Exchange.TRANSACTION_1, response_payloads, crypto=self.crypto, hashmsg=True)
    def process(self, request, stream, remote_id, reply):
        request.parse_payloads(stream, crypto=self.crypto)
        print(repr(request))
        if remote_id not in self.all_child_sa:
            self.all_child_sa[remote_id] = self.child_sa
        elif self.all_child_sa[remote_id] != self.child_sa:
            self.child_sa = self.all_child_sa[remote_id]
        if request.exchange == enums.Exchange.IDENTITY_1 and request.get_payload(enums.Payload.SA_1):
            assert self.state == State.INITIAL
            request_payload_sa = request.get_payload(enums.Payload.SA_1)
            self.sa_bytes = request_payload_sa.to_bytes()
            self.transform = request_payload_sa.proposals[0].transforms[0].values
            self.auth_mode = self.transform[enums.TransformAttr.AUTH]
            del request_payload_sa.proposals[0].transforms[1:]
            response_payloads = request.payloads
            reply(self.response(enums.Exchange.IDENTITY_1, response_payloads))
            self.state = State.SA_SENT
        elif request.exchange == enums.Exchange.IDENTITY_1 and request.get_payload(enums.Payload.KE_1):
            assert self.state == State.SA_SENT
            self.peer_public_key = request.get_payload(enums.Payload.KE_1).ke_data
            self.my_public_key, self.shared_secret = crypto.DiffieHellman(self.transform[enums.TransformAttr.DH], self.peer_public_key)
            self.peer_nonce = request.get_payload(enums.Payload.NONCE_1).nonce
            response_payloads = [ message.PayloadKE_1(self.my_public_key), message.PayloadNONCE_1(self.my_nonce),
                                  message.PayloadNATD_1(os.urandom(32)), message.PayloadNATD_1(os.urandom(32)) ]
            cipher = crypto.Cipher(self.transform[enums.TransformAttr.ENCR], self.transform[enums.TransformAttr.KEY_LENGTH])
            prf = crypto.Prf(self.transform[enums.TransformAttr.HASH])
            self.skeyid = prf.prf(self.args.passwd.encode(), self.peer_nonce+self.my_nonce)
            self.skeyid_d = prf.prf(self.skeyid, self.shared_secret+self.peer_spi+self.my_spi+bytes([0]))
            self.skeyid_a = prf.prf(self.skeyid, self.skeyid_d+self.shared_secret+self.peer_spi+self.my_spi+bytes([1]))
            self.skeyid_e = prf.prf(self.skeyid, self.skeyid_a+self.shared_secret+self.peer_spi+self.my_spi+bytes([2]))
            iv = prf.hasher(self.peer_public_key+self.my_public_key).digest()[:cipher.block_size]
            self.crypto = crypto.Crypto(cipher, self.skeyid_e[:cipher.key_size], prf=prf, iv=iv)
            reply(self.response(enums.Exchange.IDENTITY_1, response_payloads))
            self.state = State.KE_SENT
        elif request.exchange == enums.Exchange.IDENTITY_1 and request.get_payload(enums.Payload.ID_1):
            assert self.state == State.KE_SENT
            payload_id = request.get_payload(enums.Payload.ID_1)
            prf = self.crypto.prf
            hash_i = prf.prf(self.skeyid, self.peer_public_key+self.my_public_key+self.peer_spi+self.my_spi+self.sa_bytes+payload_id.to_bytes())
            assert hash_i == request.get_payload(enums.Payload.HASH_1).data, 'Authentication Failed'
            response_payload_id = message.PayloadID_1(enums.IDType.ID_FQDN, f'{__title__}-{__version__}'.encode())
            hash_r = prf.prf(self.skeyid, self.my_public_key+self.peer_public_key+self.my_spi+self.peer_spi+self.sa_bytes+response_payload_id.to_bytes())
            response_payloads = [response_payload_id, message.PayloadHASH_1(hash_r)]
            reply(self.response(enums.Exchange.IDENTITY_1, response_payloads, crypto=self.crypto))
            self.state = State.HASH_SENT
            if self.auth_mode == enums.AuthId_1.XAUTHInitPreShared:
                reply(self.xauth_init())
        elif request.exchange == enums.Exchange.TRANSACTION_1:
            self.verify_hash(request)
            payload_cp = request.get_payload(enums.Payload.CP_1)
            if enums.CPAttrType.XAUTH_USER_NAME in payload_cp.attrs:
                assert self.state == State.HASH_SENT
                response_payloads = [ message.PayloadCP_1(enums.CFGType.CFG_SET, {enums.CPAttrType.XAUTH_STATUS: 1}) ]
                self.state = State.AUTH_SET
            elif enums.CPAttrType.INTERNAL_IP4_ADDRESS in payload_cp.attrs:
                assert self.state == State.AUTH_SET
                attrs = { enums.CPAttrType.INTERNAL_IP4_ADDRESS: ipaddress.ip_address('10.0.0.1').packed,
                          enums.CPAttrType.INTERNAL_IP4_DNS: ipaddress.ip_address(self.args.dns).packed,
                        }
                response_payloads = [ message.PayloadCP_1(enums.CFGType.CFG_REPLY, attrs, identifier=payload_cp.identifier) ]
                self.state = State.CONF_SENT
            elif payload_cp.cftype == enums.CFGType.CFG_ACK:
                return
            else:
                raise Exception('Unknown CP Exchange')
            reply(self.response(enums.Exchange.TRANSACTION_1, response_payloads, request.message_id, crypto=self.crypto, hashmsg=True))
        elif request.exchange == enums.Exchange.QUICK_1 and len(request.payloads) == 1:
            assert request.payloads[0].type == enums.Payload.HASH_1
            assert self.state == State.CHILD_SA_SENT
            self.state = State.ESTABLISHED
        elif request.exchange == enums.Exchange.QUICK_1:
            assert self.state == State.CONF_SENT and self.auth_mode == enums.AuthId_1.XAUTHInitPreShared or \
                   self.state == State.HASH_SENT and self.auth_mode == enums.AuthId_1.PSK or \
                   self.child_sa
            self.verify_hash(request)
            payload_nonce = request.get_payload(enums.Payload.NONCE_1)
            peer_nonce = payload_nonce.nonce
            payload_nonce.nonce = my_nonce = os.urandom(len(peer_nonce))
            chosen_proposal = request.get_payload(enums.Payload.SA_1).proposals[0]
            del chosen_proposal.transforms[1:]
            peer_spi = chosen_proposal.spi
            chosen_proposal.spi = my_spi = os.urandom(4)
            reply(self.response(enums.Exchange.QUICK_1, request.payloads, request.message_id, crypto=self.crypto, hashmsg=peer_nonce))

            transform = chosen_proposal.transforms[0].values
            cipher = crypto.Cipher(chosen_proposal.transforms[0].id, transform[enums.ESPAttr.KEY_LENGTH])
            integ = crypto.Integrity(transform[enums.ESPAttr.AUTH])
            keymat_fmt = struct.Struct('>{0}s{1}s'.format(cipher.key_size, integ.key_size))
            keymat = self.crypto.prf.prfplus(self.skeyid_d, bytes([chosen_proposal.protocol])+my_spi+peer_nonce+my_nonce, False)
            sk_ei, sk_ai = keymat_fmt.unpack(bytes(next(keymat) for _ in range(keymat_fmt.size)))
            keymat = self.crypto.prf.prfplus(self.skeyid_d, bytes([chosen_proposal.protocol])+peer_spi+peer_nonce+my_nonce, False)
            sk_er, sk_ar = keymat_fmt.unpack(bytes(next(keymat) for _ in range(keymat_fmt.size)))
            crypto_in = crypto.Crypto(cipher, sk_ei, integ, sk_ai)
            crypto_out = crypto.Crypto(cipher, sk_er, integ, sk_ar)
            child_sa = ChildSa(my_spi, peer_spi, crypto_in, crypto_out)
            self.sessions[my_spi] = child_sa
            for old_child_sa in self.child_sa:
                old_child_sa.child = child_sa
            self.child_sa.append(child_sa)
            self.state = State.CHILD_SA_SENT
        elif request.exchange == enums.Exchange.INFORMATIONAL_1:
            self.verify_hash(request)
            response_payloads = []
            delete_payload = request.get_payload(enums.Payload.DELETE_1)
            notify_payload = request.get_payload(enums.Payload.NOTIFY_1)
            if not request.payloads:
                pass
            elif delete_payload and delete_payload.protocol == enums.Protocol.IKE:
                self.state = State.DELETED
                self.sessions.pop(self.my_spi)
                response_payloads.append(delete_payload)
                message_id = request.message_id
            elif delete_payload:
                spis = []
                for spi in delete_payload.spis:
                    child_sa = next((x for x in self.child_sa if x.spi_out == spi), None)
                    if child_sa:
                        self.child_sa.remove(child_sa)
                        self.sessions.pop(child_sa.spi_in)
                        spis.append(child_sa.spi_in)
                response_payloads.append(message.PayloadDELETE_1(delete_payload.doi, delete_payload.protocol, spis))
                message_id = request.message_id
            elif notify_payload and notify_payload.notify == enums.Notify.ISAKMP_NTYPE_R_U_THERE:
                notify_payload.notify = enums.Notify.ISAKMP_NTYPE_R_U_THERE_ACK
                response_payloads.append(notify_payload)
                message_id = request.message_id
                message_id = None
            elif notify_payload and notify_payload.notify == enums.Notify.INITIAL_CONTACT_1:
                notify_payload.notify = enums.Notify.INITIAL_CONTACT_1
                response_payloads.append(notify_payload)
                message_id = request.message_id
            else:
                raise Exception(f'unhandled informational {request!r}')
            reply(self.response(enums.Exchange.INFORMATIONAL_1, response_payloads, message_id, crypto=self.crypto, hashmsg=True))
            #if notify_payload and notify_payload.notify == enums.Notify.INITIAL_CONTACT_1:
            #    reply(self.xauth_init())
        else:
            raise Exception(f'unhandled request {request!r}')

class IKEv2Session:
    def __init__(self, args, sessions, peer_spi):
        self.args = args
        self.sessions = sessions
        self.my_spi = os.urandom(8)
        self.peer_spi = peer_spi
        self.peer_msgid = 0
        self.my_crypto = None
        self.peer_crypto = None
        self.my_nonce = os.urandom(32)
        self.peer_nonce = None
        self.state = State.INITIAL
        self.request_data = None
        self.response_data = None
        self.child_sa = []
        self.sessions[self.my_spi] = self
    def create_key(self, ike_proposal, shared_secret, old_sk_d=None):
        prf = crypto.Prf(ike_proposal.get_transform(enums.Transform.PRF).id)
        integ = crypto.Integrity(ike_proposal.get_transform(enums.Transform.INTEG).id)
        cipher = crypto.Cipher(ike_proposal.get_transform(enums.Transform.ENCR).id,
                               ike_proposal.get_transform(enums.Transform.ENCR).keylen)
        if not old_sk_d:
            skeyseed = prf.prf(self.peer_nonce+self.my_nonce, shared_secret)
        else:
            skeyseed = prf.prf(old_sk_d, shared_secret+self.peer_nonce+self.my_nonce)
        keymat_fmt = struct.Struct('>{0}s{1}s{1}s{2}s{2}s{0}s{0}s'.format(prf.key_size, integ.key_size, cipher.key_size))
        keymat = prf.prfplus(skeyseed, self.peer_nonce+self.my_nonce+self.peer_spi+self.my_spi)
        self.sk_d, sk_ai, sk_ar, sk_ei, sk_er, sk_pi, sk_pr = keymat_fmt.unpack(bytes(next(keymat) for _ in range(keymat_fmt.size)))
        self.my_crypto = crypto.Crypto(cipher, sk_er, integ, sk_ar, prf, sk_pr)
        self.peer_crypto = crypto.Crypto(cipher, sk_ei, integ, sk_ai, prf, sk_pi)
    def create_child_key(self, child_proposal, nonce_i, nonce_r):
        integ = crypto.Integrity(child_proposal.get_transform(enums.Transform.INTEG).id)
        cipher = crypto.Cipher(child_proposal.get_transform(enums.Transform.ENCR).id,
                               child_proposal.get_transform(enums.Transform.ENCR).keylen)
        keymat_fmt = struct.Struct('>{0}s{1}s{0}s{1}s'.format(cipher.key_size, integ.key_size))
        keymat = self.my_crypto.prf.prfplus(self.sk_d, nonce_i+nonce_r)
        sk_ei, sk_ai, sk_er, sk_ar = keymat_fmt.unpack(bytes(next(keymat) for _ in range(keymat_fmt.size)))
        crypto_in = crypto.Crypto(cipher, sk_ei, integ, sk_ai)
        crypto_out = crypto.Crypto(cipher, sk_er, integ, sk_ar)
        child_sa = ChildSa(os.urandom(4), child_proposal.spi, crypto_in, crypto_out)
        self.child_sa.append(child_sa)
        self.sessions[child_sa.spi_in] = child_sa
        return child_sa
    def auth_data(self, message_data, nonce, payload, sk_p):
        prf = self.peer_crypto.prf.prf
        return prf(prf(self.args.passwd.encode(), b'Key Pad for IKEv2'), message_data+nonce+prf(sk_p, payload.to_bytes()))
    def response(self, exchange, payloads, *, crypto=None):
        response = message.Message(self.peer_spi, self.my_spi, 0x20, exchange,
                                   enums.MsgFlag.Response, self.peer_msgid, payloads)
        #print(repr(response))
        self.peer_msgid += 1
        self.response_data = response.to_bytes(crypto=crypto)
        return self.response_data
    def process(self, request, stream, remote_id, reply):
        if request.message_id == self.peer_msgid - 1:
            reply(self.response_data)
            return
        elif request.message_id != self.peer_msgid:
            return
        request.parse_payloads(stream, crypto=self.peer_crypto)
        print(repr(request))
        if request.exchange == enums.Exchange.IKE_SA_INIT:
            assert self.state == State.INITIAL
            self.peer_nonce = request.get_payload(enums.Payload.NONCE).nonce
            chosen_proposal = request.get_payload(enums.Payload.SA).get_proposal(enums.EncrId.ENCR_AES_CBC)
            payload_ke = request.get_payload(enums.Payload.KE)
            prefered_dh = chosen_proposal.get_transform(enums.Transform.DH).id
            if payload_ke.dh_group != prefered_dh or payload_ke.ke_data[0] == 0:
                reply(self.response(enums.Exchange.IKE_SA_INIT, [ message.PayloadNOTIFY(0, enums.Notify.INVALID_KE_PAYLOAD, b'', prefered_dh.to_bytes(2, 'big'))]))
                return
            public_key, shared_secret = crypto.DiffieHellman(payload_ke.dh_group, payload_ke.ke_data)
            self.create_key(chosen_proposal, shared_secret)
            response_payloads = [ message.PayloadSA([chosen_proposal]),
                                  message.PayloadNONCE(self.my_nonce),
                                  message.PayloadKE(payload_ke.dh_group, public_key),
                                  message.PayloadNOTIFY(0, enums.Notify.NAT_DETECTION_DESTINATION_IP, b'', os.urandom(20)),
                                  message.PayloadNOTIFY(0, enums.Notify.NAT_DETECTION_SOURCE_IP, b'', os.urandom(20)) ]
            reply(self.response(enums.Exchange.IKE_SA_INIT, response_payloads))
            self.state = State.SA_SENT
            self.request_data = stream.getvalue()
        elif request.exchange == enums.Exchange.IKE_AUTH:
            assert self.state == State.SA_SENT
            request_payload_idi = request.get_payload(enums.Payload.IDi)
            request_payload_auth = request.get_payload(enums.Payload.AUTH)
            if request_payload_auth is None:
                EAP = True
                raise Exception('EAP not supported')
            else:
                EAP = False
                auth_data = self.auth_data(self.request_data, self.my_nonce, request_payload_idi, self.peer_crypto.sk_p)
                assert auth_data == request_payload_auth.auth_data, 'Authentication Failed'
            chosen_child_proposal = request.get_payload(enums.Payload.SA).get_proposal(enums.EncrId.ENCR_AES_CBC)
            child_sa = self.create_child_key(chosen_child_proposal, self.peer_nonce, self.my_nonce)
            chosen_child_proposal.spi = child_sa.spi_in
            response_payload_idr = message.PayloadIDr(enums.IDType.ID_FQDN, f'{__title__}-{__version__}'.encode())
            auth_data = self.auth_data(self.response_data, self.peer_nonce, response_payload_idr, self.my_crypto.sk_p)

            response_payloads = [ message.PayloadSA([chosen_child_proposal]),
                                  request.get_payload(enums.Payload.TSi),
                                  request.get_payload(enums.Payload.TSr),
                                  response_payload_idr,
                                  message.PayloadAUTH(enums.AuthMethod.PSK, auth_data) ]
            if request.get_payload(enums.Payload.CP):
                attrs = { enums.CPAttrType.INTERNAL_IP4_ADDRESS: ipaddress.ip_address('1.0.0.1').packed,
                          enums.CPAttrType.INTERNAL_IP4_DNS: ipaddress.ip_address(self.args.dns).packed, }
                response_payloads.append(message.PayloadCP(enums.CFGType.CFG_REPLY, attrs))
            reply(self.response(enums.Exchange.IKE_AUTH, response_payloads, crypto=self.my_crypto))
            self.state = State.ESTABLISHED
        elif request.exchange == enums.Exchange.INFORMATIONAL:
            assert self.state == State.ESTABLISHED
            response_payloads = []
            delete_payload = request.get_payload(enums.Payload.DELETE)
            if not request.payloads:
                pass
            elif delete_payload and delete_payload.protocol == enums.Protocol.IKE:
                self.state = State.DELETED
                self.sessions.pop(self.my_spi)
                for child_sa in self.child_sa:
                    self.sessions.pop(child_sa.spi_in)
                self.child_sa = []
                response_payloads.append(delete_payload)
            elif delete_payload:
                spis = []
                for spi in delete_payload.spis:
                    child_sa = next((x for x in self.child_sa if x.spi_out == spi), None)
                    if child_sa:
                        self.child_sa.remove(child_sa)
                        self.sessions.pop(child_sa.spi_in)
                        spis.append(child_sa.spi_in)
                response_payloads.append(message.PayloadDELETE(delete_payload.protocol, spis))
            else:
                raise Exception(f'unhandled informational {request!r}')
            reply(self.response(enums.Exchange.INFORMATIONAL, response_payloads, crypto=self.my_crypto))
        elif request.exchange == enums.Exchange.CREATE_CHILD_SA:
            assert self.state == State.ESTABLISHED
            chosen_proposal = request.get_payload(enums.Payload.SA).get_proposal(enums.EncrId.ENCR_AES_CBC)
            if chosen_proposal.protocol != enums.Protocol.IKE:
                payload_notify = request.get_payload_notify(enums.Notify.REKEY_SA)
                if not payload_notify:
                    raise Exception(f'unhandled protocol {chosen_proposal.protocol} {request!r}')
                old_child_sa = next(i for i in self.child_sa if i.spi_out == payload_notify.spi)
                peer_nonce = request.get_payload(enums.Payload.NONCE).nonce
                my_nonce = os.urandom(random.randrange(16, 256))
                child_sa = self.create_child_key(chosen_proposal, peer_nonce, my_nonce)
                chosen_proposal.spi = child_sa.spi_in
                old_child_sa.child = child_sa
                response_payloads = [ message.PayloadNOTIFY(chosen_proposal.protocol, enums.Notify.REKEY_SA, old_child_sa.spi_in, b''),
                                      message.PayloadNONCE(my_nonce),
                                      message.PayloadSA([chosen_proposal]),
                                      request.get_payload(enums.Payload.TSi),
                                      request.get_payload(enums.Payload.TSr) ]
            else:
                child = IKEv2Session(self.args, self.sessions, chosen_proposal.spi)
                child.state = State.ESTABLISHED
                child.peer_nonce = request.get_payload(enums.Payload.NONCE).nonce
                child.child_sa = self.child_sa
                self.child_sa = []
                payload_ke = request.get_payload(enums.Payload.KE)
                public_key, shared_secret = crypto.DiffieHellman(payload_ke.dh_group, payload_ke.ke_data)
                chosen_proposal.spi = child.my_spi
                child.create_key(chosen_proposal, shared_secret, self.sk_d)
                response_payloads = [ message.PayloadSA([chosen_proposal]),
                                      message.PayloadNONCE(child.my_nonce),
                                      message.PayloadKE(payload_ke.dh_group, public_key) ]
            reply(self.response(enums.Exchange.CREATE_CHILD_SA, response_payloads, crypto=self.my_crypto))
        else:
            raise Exception(f'unhandled request {request!r}')

IKE_HEADER = b'\x00\x00\x00\x00'

class IKE_500(asyncio.DatagramProtocol):
    def __init__(self, args, sessions):
        self.args = args
        self.sessions = sessions
    def connection_made(self, transport):
        self.transport = transport
    def datagram_received(self, data, addr, *, response_header=b''):
        stream = io.BytesIO(data)
        request = message.Message.parse(stream)
        if request.exchange == enums.Exchange.IKE_SA_INIT:
            session = IKEv2Session(self.args, self.sessions, request.spi_i)
        elif request.exchange == enums.Exchange.IDENTITY_1 and request.spi_r == bytes(8):
            session = IKEv1Session(self.args, self.sessions, request.spi_i)
        elif request.spi_r in self.sessions:
            session = self.sessions[request.spi_r]
        else:
            return
        session.process(request, stream, addr[:2], lambda response: self.transport.sendto(response_header+response, addr))

class SPE_4500(IKE_500):
    def __init__(self, args, sessions):
        IKE_500.__init__(self, args, sessions)
        self.ippacket = ip.IPPacket(args)
    def datagram_received(self, data, addr):
        spi = data[:4]
        if spi == b'\xff':
            self.transport.sendto(b'\xff', addr)
        elif spi == IKE_HEADER:
            IKE_500.datagram_received(self, data[4:], addr, response_header=IKE_HEADER)
        elif spi in self.sessions:
            sa = self.sessions[spi]
            seqnum = int.from_bytes(data[4:8], 'big')
            if seqnum < sa.msgid_in or seqnum in sa.msgwin_in:
                return
            if sa.msgid_in == 1 and sa.crypto_in.integrity.hasher is hashlib.sha256 and (len(data)-8)%16 == 12:
                # HMAC-SHA2-256-96 fix
                sa.crypto_in.integrity.hash_size = 12
                sa.crypto_out.integrity.hash_size = 12
            sa.crypto_in.verify_checksum(data)
            if seqnum > sa.msgid_in + 65536:
                sa.incr_msgid_in()
            if seqnum == sa.msgid_in:
                sa.incr_msgid_in()
            else:
                sa.msgwin_in.add(seqnum)
            header, data = sa.crypto_in.decrypt_esp(data[8:])
            def reply(data):
                nonlocal sa
                while sa and sa.spi_in not in self.sessions:
                    sa = sa.child
                if not sa:
                    return False
                encrypted = bytearray(sa.crypto_out.encrypt_esp(header, data))
                encrypted[0:0] = sa.spi_out + sa.msgid_out.to_bytes(4, 'big')
                sa.crypto_out.add_checksum(encrypted)
                sa.msgid_out += 1
                self.transport.sendto(encrypted, addr)
                return True
            self.ippacket.handle(addr[:2], header, data, reply)
        else:
            print('unknown packet', data, addr)

class WIREGUARD(asyncio.DatagramProtocol):
    def __init__(self, args):
        self.args = args
        self.preshared_key = b'\x00'*32
        self.ippacket = ip.IPPacket(args)
        self.private_key = hashlib.blake2s(args.passwd.encode()).digest()
        self.public_key = crypto.X25519(self.private_key, 9)
        self.keys = {}
        self.index_generators = {}
        self.sender_index_generator = itertools.count()
        print('======== WIREGUARD SETTING ========')
        print('PublicKey:', base64.b64encode(self.public_key).decode())
        print('===================================')
    def connection_made(self, transport):
        self.transport = transport
    def datagram_received(self, data, addr):
        cmd = int.from_bytes(data[0:4], 'little')
        if cmd == 1 and len(data) == 148:
            HASH = lambda x: hashlib.blake2s(x).digest()
            MAC  = lambda key, x: hashlib.blake2s(x, key=key, digest_size=16).digest()
            HMAC = lambda key, x: hmac.digest(key, x, hashlib.blake2s)
            p, mac1, mac2 = struct.unpack('<116s16s16s', data)
            assert mac1 == MAC(HASH(b"mac1----" + self.public_key), p)
            assert mac2 == b'\x00'*16
            index = next(self.sender_index_generator)
            sender_index, unencrypted_ephemeral, encrypted_static, encrypted_timestamp = struct.unpack('<4xI32s48s28s', data[:-32])

            chaining_key = HASH(b"Noise_IKpsk2_25519_ChaChaPoly_BLAKE2s")
            hash0 = HASH(HASH(HASH(chaining_key + b"WireGuard v1 zx2c4 Jason@zx2c4.com") + self.public_key) + unencrypted_ephemeral)
            chaining_key = HMAC(HMAC(chaining_key, unencrypted_ephemeral), b"\x01")
            temp = HMAC(chaining_key, crypto.X25519(self.private_key, unencrypted_ephemeral))
            chaining_key = HMAC(temp, b"\x01")
            static_public = crypto.aead_chacha20poly1305_decrypt(HMAC(temp, chaining_key + b"\x02"), 0, encrypted_static, hash0)
            hash0 = HASH(hash0 + encrypted_static)
            temp = HMAC(chaining_key, crypto.X25519(self.private_key, static_public))
            chaining_key = HMAC(temp, b"\x01")
            timestamp = crypto.aead_chacha20poly1305_decrypt(HMAC(temp, chaining_key + b"\x02"), 0, encrypted_timestamp, hash0)
            hash0 = HASH(hash0 + encrypted_timestamp)

            ephemeral_private = os.urandom(32)
            ephemeral_public = crypto.X25519(ephemeral_private, 9)
            hash0 = HASH(hash0 + ephemeral_public)
            chaining_key = HMAC(HMAC(HMAC(HMAC(HMAC(HMAC(chaining_key, ephemeral_public), b"\x01"), crypto.X25519(ephemeral_private, unencrypted_ephemeral)), b"\x01"), crypto.X25519(ephemeral_private, static_public)), b"\x01")
            temp = HMAC(chaining_key, self.preshared_key)
            chaining_key = HMAC(temp, b"\x01")
            temp2 = HMAC(temp, chaining_key + b"\x02")
            key = HMAC(temp, temp2 + b"\x03")
            hash0 = HASH(hash0 + temp2)
            encrypted_nothing = crypto.aead_chacha20poly1305_encrypt(key, 0, b"", hash0)
            #hash0 = HASH(hash0 + encrypted_nothing)
            msg = struct.pack('<III32s16s', 2, index, sender_index, ephemeral_public, encrypted_nothing)
            msg = msg + MAC(HASH(b"mac1----" + static_public), msg) + b'\x00'*16
            self.transport.sendto(msg, addr)
            print('login', addr, sender_index)

            temp1 = HMAC(chaining_key, b"")
            temp2 = HMAC(temp1, b"\x01")
            temp3 = HMAC(temp1, temp2 + b"\x02")
            self.keys[index] = [sender_index, temp2, temp3]
            self.index_generators[index] = itertools.count()
        elif cmd == 4 and len(data) >= 32:
            _, index, counter = struct.unpack('<IIQ', data[:16])
            sender_index, receiving_key, sending_key = self.keys[index]
            packet = crypto.aead_chacha20poly1305_decrypt(receiving_key, counter, data[16:], b'')
            def reply(data):
                counter = next(self.index_generators[index])
                data = data + b''*((-len(data))%16)
                msg = crypto.aead_chacha20poly1305_encrypt(sending_key, counter, data, b'')
                msg = struct.pack('<IIQ', 4, sender_index, counter) + msg
                self.transport.sendto(msg, addr)
                return True
            if packet:
                self.ippacket.handle_ipv4(addr[:2], packet, reply)
            else:
                reply(b'')


def main():
    parser = argparse.ArgumentParser(description=__description__, epilog=f'Online help: <{__url__}>')
    parser.add_argument('-wg', dest='wireguard', default=None, type=int, help='start a wireguard vpn with port number (default: off)')
    parser.add_argument('-r', dest='rserver', default=[], action='append', type=pproxy.Connection, help='tcp remote server uri (default: direct)')
    parser.add_argument('-ur', dest='urserver', default=[], action='append', type=pproxy.Connection, help='udp remote server uri (default: direct)')
    parser.add_argument('-s', dest='salgorithm', default='fa', choices=('fa', 'rr', 'rc', 'lc'), help='scheduling algorithm (default: first_available)')
    parser.add_argument('-p', dest='passwd', default='test', help='password (default: test)')
    parser.add_argument('-dns', dest='dns', default='1.1.1.1', help='dns server (default: 1.1.1.1)')
    parser.add_argument('-nc', dest='nocache', default=None, action='store_true', help='do not cache dns (default: off)')
    parser.add_argument('-v', dest='v', action='count', help='print verbose output')
    parser.add_argument('--version', action='version', version=f'{__title__} {__version__}')
    args = parser.parse_args()
    args.DIRECT = pproxy.Connection('direct://')
    loop = asyncio.get_event_loop()
    sessions = {}
    transport1, _ = loop.run_until_complete(loop.create_datagram_endpoint(lambda: IKE_500(args, sessions), ('0.0.0.0', 500)))
    transport2, _ = loop.run_until_complete(loop.create_datagram_endpoint(lambda: SPE_4500(args, sessions), ('0.0.0.0', 4500)))
    print('Serving on UDP :500 :4500...')
    if args.wireguard:
        transport3, _ = loop.run_until_complete(loop.create_datagram_endpoint(lambda: WIREGUARD(args), ('0.0.0.0', args.wireguard)))
        print(f'Serving on UDP :{args.wireguard} (WIREGUARD)...')
    else:
        transport3 = None
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        print('exit')
    for task in asyncio.all_tasks(loop) if hasattr(asyncio, 'all_tasks') else asyncio.Task.all_tasks():
        task.cancel()
    transport1.close()
    transport2.close()
    if transport3:
        transport3.close()
    loop.run_until_complete(loop.shutdown_asyncgens())
    loop.close()

if __name__ == '__main__':
    main()



__title__       = "pvpn"
__version__     = "0.2.2"
__license__     = "MIT"
__description__ = "Python VPN server."
__keywords__    = "vpn ike esp ip tcp udp wireguard"
__author__      = "Qian Wenjie"
__email__       = "qianwenjie@gmail.com"
__url__         = "https://github.com/qwj/python-vpn"

__all__ = ['__title__', '__version__', '__description__', '__url__']



import enum

class TypeEnum(enum.IntEnum):
    @classmethod
    def _missing_(cls, value):
        obj = int.__new__(cls, value)
        obj._name_ = f'{cls.__name__}_{value}'
        obj._value_ = value
        return obj

class MsgFlag(enum.IntFlag):
    NONE = 0x00
    Encryption = 0x01
    Commit = 0x02
    Authentication = 0x04
    Initiator = 0x08
    CanUseHigherVersion = 0x10
    Response = 0x20

class Exchange(TypeEnum):
    IKE_BASE_1 = 1
    IDENTITY_1 = 2
    IKE_AUTH_1 = 3
    IKE_AGGRESIVE_1 = 4
    INFORMATIONAL_1 = 5
    TRANSACTION_1 = 6
    QUICK_1 = 32
    NEW_GROUP_1 = 33
    IKE_SA_INIT = 34
    IKE_AUTH = 35
    CREATE_CHILD_SA = 36
    INFORMATIONAL = 37
    IKE_SESSION_RESUME = 38
    GSA_AUTH = 39
    GSA_REGISTRATION = 40
    GSA_REKEY = 41

class Payload(TypeEnum):
    NONE = 0
    SA_1 = 1
    PROPOSAL_1 = 2
    TRANSFORM_1 = 3
    KE_1 = 4
    ID_1 = 5
    CERT_1 = 6
    CERTREQ_1 = 7
    HASH_1 = 8
    SIG_1 = 9
    NONCE_1 = 10
    NOTIFY_1 = 11
    DELETE_1 = 12
    VENDOR_1 = 13
    CP_1 = 14
    SAK_1 = 15
    SAT_1 = 16
    KD_1 = 17
    SEQ_1 = 18
    POP_1 = 19
    NATD_1 = 20
    NATO_1 = 21
    GAP_1 = 22
    SA = 33
    KE = 34
    IDi = 35
    IDr = 36
    CERT = 37
    CERTREQ = 38
    AUTH = 39
    NONCE = 40
    NOTIFY = 41
    DELETE = 42
    VENDOR = 43
    TSi = 44
    TSr = 45
    SK = 46
    CP = 47
    EAP = 48
    GSPM = 49
    IDg = 50
    GSA = 51
    KD = 52
    SKF = 53
    PS = 54

class Protocol(TypeEnum):
    NONE = 0
    IKE = 1
    AH = 2
    ESP = 3
    FC_ESP_HEADER = 4
    FC_CT_AUTHENTICATION = 5

class Transform(TypeEnum):
    ENCR = 1
    PRF = 2
    INTEG = 3
    DH = 4
    ESN = 5

class EncrId(TypeEnum):
    ENCR_DES = 2
    ENCR_3DES = 3
    ENCR_RC5 = 4
    ENCR_IDEA = 5
    ENCR_CAST = 6
    ENCR_BLOWFISH = 7
    ENCR_3IDEA = 8
    ENCR_DES_IV32 = 9
    ENCR_NULL = 11
    ENCR_AES_CBC = 12
    ENCR_AES_CTR = 13
    ENCR_AES_CCM_8 = 14
    ENCR_AES_CCM_12 = 15
    ENCR_AES_CCM_16 = 16
    ENCR_AES_GCM_8 = 18
    ENCR_AES_GCM_12 = 19
    ENCR_AES_GCM_16 = 20
    ENCR_NULL_AUTH_AES_GMAC = 21
    ENCR_CAMELLIA_CBC = 23
    ENCR_CAMELLIA_CTR = 24
    ENCR_CAMELLIA_CCM_8 = 25
    ENCR_CAMELLIA_CCM_12 = 26
    ENCR_CAMELLIA_CCM_16 = 27
    ENCR_CHACHA20_POLY1305 = 28
    ENCR_AES_CCM_8_IIV = 29
    ENCR_AES_GCM_16_IIV = 30
    ENCR_CHACHA20_POLY1305_IIV = 31

class PrfId(TypeEnum):
    PRF_HMAC_MD5 = 1
    PRF_HMAC_SHA1 = 2
    PRF_HMAC_TIGER = 3
    PRF_AES128_XCBC = 4
    PRF_HMAC_SHA2_256 = 5
    PRF_HMAC_SHA2_384 = 6
    PRF_HMAC_SHA2_512 = 7
    PRF_AES128_CMAC = 8

class IntegId(TypeEnum):
    AUTH_NONE = 0
    AUTH_HMAC_MD5_96 = 1
    AUTH_HMAC_SHA1_96 = 2
    AUTH_DES_MAC = 3
    AUTH_KPDK_MD5 = 4
    AUTH_AES_XCBC_96 = 5
    AUTH_HMAC_MD5_128 = 6
    AUTH_HMAC_SHA1_160 = 7
    AUTH_AES_CMAC_96 = 8
    AUTH_AES_128_GMAC = 9
    AUTH_AES_192_GMAC = 10
    AUTH_AES_256_GMAC = 11
    AUTH_HMAC_SHA2_256_128 = 12
    AUTH_HMAC_SHA2_384_192 = 13
    AUTH_HMAC_SHA2_512_256 = 14

class DhId(TypeEnum):
    DH_NONE = 0
    DH_1 = 1
    DH_2 = 2
    DH_5 = 5
    DH_14 = 14
    DH_15 = 15
    DH_16 = 16
    DH_17 = 17
    DH_18 = 18
    DH_19 = 19
    DH_20 = 20
    DH_21 = 21
    DH_22 = 22
    DH_23 = 23
    DH_24 = 24
    DH_25 = 25
    DH_26 = 26
    DH_27 = 27
    DH_28 = 28
    DH_29 = 29
    DH_30 = 30
    DH_31 = 31
    DH_32 = 32

class EsnId(TypeEnum):
    NO_ESN = 0
    ESN = 1

TransformTable = {
    Transform.ENCR: EncrId,
    Transform.PRF: PrfId,
    Transform.INTEG: IntegId,
    Transform.DH: DhId,
    Transform.ESN: EsnId
}

class Notify(TypeEnum):
    UNSUPPORTED_CRITICAL_PAYLOAD = 1
    DOI_NOT_SUPPORTED = 2
    SITUATION_NOT_SUPPORTED = 3
    INVALID_IKE_SPI = 4
    INVALID_MAJOR_VERSION = 5
    INVALID_MINOR_VERSION = 6
    INVALID_SYNTAX = 7
    INVALID_FLAGS = 8
    INVALID_MESSAGE_ID = 9
    INVALID_PROTOCOL_ID = 10
    INVALID_SPI = 11
    INVALID_TRANSFORM_ID = 12
    ATTRIBUTES_NOT_SUPPORTED = 13
    NO_PROPOSAL_CHOSEN = 14
    BAD_PROPOSAL_SYNTAX = 15
    PAYLOAD_MALFORMED = 16
    INVALID_KE_PAYLOAD = 17
    INVALID_ID_INFORMATION = 18
    INVALID_CERT_ENCODING = 19
    INVALID_CERTIFICATE = 20
    CERT_TYPE_UNSUPPORTED = 21
    INVALID_CERT_AUTHORITY = 22
    INVALID_HASH_INFORMATION = 23
    AUTHENTICATION_FAILED = 24
    INVALID_SIGNATURE = 25
    ADDRESS_NOTIFICATION = 26
    NOTIFY_SA_LIFETIME = 27
    CERTIFICATE_UNAVAILABLE = 28
    UNSUPPORTED_EXCHANGE_TYPE = 29
    UNEQUAL_PAYLOAD_LENGTHS = 30
    SINGLE_PAIR_REQUIRED = 34
    NO_ADDITIONAL_SAS = 35
    INTERNAL_ADDRESS_FAILURE = 36
    FAILED_CP_REQUIRED = 37
    TS_UNACCEPTABLE = 38
    INVALID_SELECTORS = 39
    UNACCEPTABLE_ADDRESSES = 40
    UNEXPECTED_NAT_DETECTED = 41
    USE_ASSIGNED_HoA = 42
    TEMPORARY_FAILURE = 43
    CHILD_SA_NOT_FOUND = 44
    INVALID_GROUP_ID = 45
    AUTHORIZATION_FAILED = 46
    INITIAL_CONTACT = 16384
    SET_WINDOW_SIZE = 16385
    ADDITIONAL_TS_POSSIBLE = 16386
    IPCOMP_SUPPORTED = 16387
    NAT_DETECTION_SOURCE_IP = 16388
    NAT_DETECTION_DESTINATION_IP = 16389
    COOKIE = 16390
    USE_TRANSPORT_MODE = 16391
    HTTP_CERT_LOOKUP_SUPPORTED = 16392
    REKEY_SA = 16393
    ESP_TFC_PADDING_NOT_SUPPORTED = 16394
    NON_FIRST_FRAGMENTS_ALSO = 16395
    MOBIKE_SUPPORTED = 16396
    ADDITIONAL_IP4_ADDRESS = 16397
    ADDITIONAL_IP6_ADDRESS = 16398
    NO_ADDITIONAL_ADDRESSES = 16399
    UPDATE_SA_ADDRESSES = 16400
    COOKIE2 = 16401
    NO_NATS_ALLOWED = 16402
    AUTH_LIFETIME = 16403
    MULTIPLE_AUTH_SUPPORTED = 16404
    ANOTHER_AUTH_FOLLOWS = 16405
    REDIRECT_SUPPORTED = 16406
    REDIRECT = 16407
    REDIRECTED_FROM = 16408
    TICKET_LT_OPAQUE = 16409
    TICKET_REQUEST = 16410
    TICKET_ACK = 16411
    TICKET_NACK = 16412
    TICKET_OPAQUE = 16413
    LINK_ID = 16414
    USE_WESP_MODE = 16415
    ROHC_SUPPORTED = 16416
    EAP_ONLY_AUTHENTICATION = 16417
    CHILDLESS_IKEV2_SUPPORTED = 16418
    QUICK_CRASH_DETECTION = 16419
    IKEV2_MESSAGE_ID_SYNC_SUPPORTED = 16420
    IPSEC_REPLAY_COUNTER_SYNC_SUPPORTED = 16421
    IKEV2_MESSAGE_ID_SYNC = 16422
    IPSEC_REPLAY_COUNTER_SYNC = 16423
    SECURE_PASSWORD_METHODS = 16424
    PSK_PERSIST = 16425
    PSK_CONFIRM = 16426
    ERX_SUPPORTED = 16427
    IFOM_CAPABILITY = 16428
    SENDER_REQUEST_ID = 16429
    IKEV2_FRAGMENTATION_SUPPORTED = 16430
    SIGNATURE_HASH_ALGORITHMS = 16431
    CLONE_IKE_SA_SUPPORTED = 16432
    CLONE_IKE_SA = 16433
    PUZZLE = 16434
    USE_PPK = 16435
    PPK_IDENTITY = 16436
    NO_PPK_AUTH = 16437
    RESPONDER_LIFETIME = 24576
    REPLAY_STATUS = 24577
    INITIAL_CONTACT_1 = 24578
    ISAKMP_NTYPE_R_U_THERE = 36136
    ISAKMP_NTYPE_R_U_THERE_ACK = 36137
    ISAKMP_NTYPE_LOAD_BALANCE = 40501
    ISAKMP_NTYPE_HEARTBEAT = 40503

class IDType(TypeEnum):
    ID_IPV4_ADDR = 1
    ID_FQDN = 2
    ID_RFC822_ADDR = 3
    ID_IPV4_ADDR_SUBNET = 4
    ID_IPV6_ADDR = 5
    ID_IPV6_ADDR_SUBNET = 6
    ID_IPV4_ADDR_RANGE = 7
    ID_IPV6_ADDR_RANGE = 8
    ID_DER_ASN1_DN = 9
    ID_DER_ASN1_GN = 10
    ID_KEY_ID = 11
    ID_FC_NAME = 12
    ID_NULL = 13

class AuthMethod(TypeEnum):
    RSA = 1
    PSK = 2
    DSS = 3
    ECDSA_SHA_256 = 9
    ECDSA_SHA_384 = 10
    ECDSA_SHA_512 = 11
    GENERIC = 12
    NULL_AUTH = 13
    DIGITAL = 14

class CFGType(TypeEnum):
    CFG_REQUEST = 1
    CFG_REPLY = 2
    CFG_SET = 3
    CFG_ACK = 4

class CPAttrType(TypeEnum):
    INTERNAL_IP4_ADDRESS = 1
    INTERNAL_IP4_NETMASK = 2
    INTERNAL_IP4_DNS = 3
    INTERNAL_IP4_NBNS = 4
    INTERNAL_ADDRESS_EXPIRY = 5
    INTERNAL_IP4_DHCP = 6
    APPLICATION_VERSION = 7
    INTERNAL_IP6_ADDRESS = 8
    INTERNAL_IP6_DNS = 10
    INTERNAL_IP6_DHCP = 12
    INTERNAL_IP4_SUBNET = 13
    SUPPORTED_ATTRIBUTES = 14
    INTERNAL_IP6_SUBNET = 15
    MIP6_HOME_PREFIX = 16
    INTERNAL_IP6_LINK = 17
    INTERNAL_IP6_PREFIX = 18
    HOME_AGENT_ADDRESS = 19
    P_CSCF_IP4_ADDRESS = 20
    P_CSCF_IP6_ADDRESS = 21
    FTT_KAT = 22
    EXTERNAL_SOURCE_IP4_NAT_INFO = 23
    TIMEOUT_PERIOD_FOR_LIVENESS_CHECK = 24
    INTERNAL_DNS_DOMAIN = 25
    INTERNAL_DNSSEC_TA = 26
    XAUTH_TYPE = 16520
    XAUTH_USER_NAME = 16521
    XAUTH_USER_PASSWORD = 16522
    XAUTH_PASSCODE = 16523
    XAUTH_MESSAGE = 16524
    XAUTH_CHALLENGE = 16525
    XAUTH_DOMAIN = 16526
    XAUTH_STATUS = 16527
    UNITY_BANNER = 28672
    UNITY_SAVE_PASSWD = 28673
    UNITY_DEF_DOMAIN = 28674
    UNITY_SPLITDNS_NAME = 28675
    UNITY_SPLIT_INCLUDE = 28676
    UNITY_NATT_PORT = 28677
    UNITY_LOCAL_LAN = 28678
    UNITY_PFS = 28679
    UNITY_FW_TYPE = 28680
    UNITY_BACKUP_SERVERS = 28681
    UNITY_DDNS_HOSTNAME = 28682
    CICSO_UNKNOWN_SEEN_ON_IPHONE = 28683

class TSType(TypeEnum):
    TS_IPV4_ADDR_RANGE = 7
    TS_IPV6_ADDR_RANGE = 8
    TS_FC_ADDR_RANGE = 9

class IpProto(TypeEnum):
    ANY = 0
    ICMP = 1
    IGMP = 2
    GGP = 3
    IPV4 = 4
    TCP = 6
    UDP = 17
    RDP = 27
    IPV6 = 41
    ESP = 50
    ICMPV6 = 58
    MH = 135
    RAW = 255

class EAPCode(TypeEnum):
    REQUEST = 1
    RESPONSE = 2
    SUCCESS = 3
    FAILURE = 4
    INITIATE = 5
    FINISH = 6

class TransformAttr(TypeEnum):
    ENCR = 1
    HASH = 2
    AUTH = 3
    DH = 4
    DH_TYPE = 5
    DH_PRIME = 6
    GENERATOR_1 = 7
    GENERATOR_2 = 8
    CURVE_A = 9
    CURVE_B = 10
    LIFETYPE = 11
    DURATION = 12
    PRF = 13
    KEY_LENGTH = 14
    FIELD_SIZE = 15
    DH_ORDER = 16

class ESPAttr(TypeEnum):
    LIFE_TYPE = 1
    DURATION = 2
    GRP_DESC = 3
    ENC_MODE = 4
    AUTH = 5
    KEY_LENGTH = 6
    KEY_ROUND = 7
    COMP_DICT_SIZE = 8
    COMP_PRIVALG = 9
    SECCTX = 10
    ESN = 11
    AUTH_KEY_LENGTH = 12
    SIG_ALGORITHM = 13
    ADDR_PRESERVE = 14
    SA_DIRECTION = 15

class EncModeId_1(TypeEnum):
    ANY = 0
    TUNNEL = 1
    TRNS = 2
    UDPTUNNEL_RFC = 3
    UDPTRNS_RFC = 4
    UDPTUNNEL_DRAFT = 61443
    UDPTRNS_DRAFT = 61444

class IntegId_1(TypeEnum):
    AUTH_NONE = 0
    AUTH_HMAC_MD5 = 1
    AUTH_HMAC_SHA1 = 2
    AUTH_DES_MAC = 3
    AUTH_KPDK = 4
    AUTH_HMAC_SHA2_256 = 5
    AUTH_HMAC_SHA2_384 = 6
    AUTH_HMAC_SHA2_512 = 7
    AUTH_HMAC_RIPEMD = 8
    AUTH_AES_XCBC_MAC = 9
    AUTH_SIG_RSA = 10
    AUTH_AES_128_GMAC = 11
    AUTH_AES_192_GMAC = 12
    AUTH_AES_256_GMAC = 13

ESPTable_1 = {
    ESPAttr.ENC_MODE: EncModeId_1,
    ESPAttr.AUTH: IntegId_1,
}

class EncrId_1(TypeEnum):
    DES_CBC = 1
    IDEA_CBC = 2
    BLOWFISH_CBC = 3
    RC5_R16_B64_CBC = 4
    _3DES_CBC = 5
    CAST_CBC = 6
    AES_CBC = 7
    CAMELLIA_CBC = 8

class HashId_1(TypeEnum):
    MD5 = 1
    SHA = 2
    TIGER = 3
    SHA2_256 = 4
    SHA2_384 = 5
    SHA2_512 = 6

class AuthId_1(TypeEnum):
    PSK = 1
    DSS = 2
    RSA = 3
    ENCR_RSA = 4
    RE_ENCR_RSA = 5
    ECDSA_SHA_256 = 9
    ECDSA_SHA_384 = 10
    ECDSA_SHA_512 = 11
    XAUTHInitPreShared = 65001
    XAUTHRespPreShared = 65002
    XAUTHInitDSS       = 65003
    XAUTHRespDSS       = 65004
    XAUTHInitRSA       = 65005
    XAUTHRespRSA       = 65006
    XAUTHInitRSAEncryption = 65007
    XAUTHRespRSAEncryption = 65008
    XAUTHInitRSARevisedEncryption = 65009
    XAUTHRespRSARevisedEncryption = 65010

TransformTable_1 = {
    TransformAttr.ENCR: EncrId_1,
    TransformAttr.HASH: HashId_1,
    TransformAttr.AUTH: AuthId_1,
    TransformAttr.DH: DhId,
}

class L2TPType(TypeEnum):
    SCCRQ = 1
    SCCRP = 2
    SCCCN = 3
    StopCCN = 4
    HELLO = 6
    OCRQ = 7
    OCRP = 8
    OCCN = 9
    ICRQ = 10
    ICRP = 11
    ICCN = 12
    CDN = 14
    WEN = 15
    SLI = 16

class L2TPAttr(TypeEnum):
    MsgType = 0
    RandomVector = 36
    Result = 1
    Version = 2
    FramingCap = 3
    BearerCap = 4
    TieBreaker = 5
    Firmware = 6
    HostName = 7
    VendorName = 8
    TunnelID = 9
    WindowSize = 10
    Challenge = 11
    Response = 13
    CauseCode = 12
    SessionID = 14
    CallSerial = 15
    MinimumBPS = 16
    MaximumBPS = 17
    BearerType = 18
    FramingType = 19
    CalledNumber = 21
    CallingNumber = 22
    SubAddress = 23
    ConnectSpeed = 24
    RxConnectSpeed = 38
    PhysicalChannel = 25
    PrivateGroupID = 37
    SequencingRequired = 39
    InitialLCP = 26
    LastSentLCP = 27
    LastReceivedLCP = 28
    ProxyAuthenType = 29
    ProxyAuthenName = 30
    ProxyAuthenChallenge = 31
    ProxyAuthenID = 32
    ProxyAuthenResponse = 33
    CallErrors = 34
    ACCM = 35





import struct, ipaddress, os, asyncio, random, heapq, collections, time, enum
from . import enums, dns

SMSS = 1350

def parse_ipv4(data):
    ihl = data[0]&0x0f
    length = int.from_bytes(data[2:4], 'big')
    proto = enums.IpProto(data[9])
    src_ip = ipaddress.ip_address(data[12:16])
    dst_ip = ipaddress.ip_address(data[16:20])
    body = data[ihl<<2:length]
    return proto, src_ip, dst_ip, body

def checksum(data):
    x = sum(struct.unpack(f'>{len(data)//2}H', data))
    while x > 0xffff:
        x = (x>>16)+(x&0xffff)
    x = 65535 - x
    return x.to_bytes(2, 'big')

def make_ipv4(proto, src_ip, dst_ip, body):
    ip_header = bytearray(struct.pack('>BxH2s2xBB2x4s4s', 0x45, len(body)+20, os.urandom(2), 64,
        proto, src_ip.packed, dst_ip.packed))
    ip_header[10:12] = checksum(ip_header)
    return bytes(ip_header+body)

def parse_udp(data):
    src_port, dst_port = struct.unpack('>HH', data[:4])
    return src_port, dst_port, data[8:]

def make_udp(src_port, dst_port, body):
    return struct.pack('>HHHH', src_port, dst_port, len(body)+8, 0)+body

def parse_icmp(data):
    icmptp, code = struct.unpack('>BB', data[:2])
    return icmptp, code, data[8:]

def parse_tcp(data):
    src_port, dst_port = struct.unpack('>HH', data[:4])
    offset = data[12]
    flag = data[13]
    body = data[offset>>2:]
    return src_port, dst_port, flag, body

def parse_l2tp(data):
    if data[0] & 0x80:
        length, tunnel_id, session_id, ns, nr = struct.unpack('>HHHHH', data[2:12])
        body = {}
        pos = 12
        while pos < len(data):
            avplen, vendor_id, attrtp = struct.unpack('>HHH', data[pos:pos+6])
            avplen &= 0x3ff
            value = data[pos+6:pos+avplen]
            if attrtp == 0:
                value = enums.L2TPType(struct.unpack('>H', value)[0])
            elif attrtp in (2, 9, 10, 14):
                value, = struct.unpack('>H', value)
            elif attrtp in (3, 15, 19, 24):
                value, = struct.unpack('>I', value)
            body[enums.L2TPAttr(attrtp)] = value
            pos += avplen
        return tunnel_id, session_id, ns, nr, body
    else:
        pos = 2
        if data[0] & 0x40:
            pos += 2
        tunnel_id, session_id = struct.unpack('>HH', data[pos:pos+4])
        pos += 4
        if data[0] & 0x08:
            ns, nr = struct.unpack('>HH', data[pos:pos+4])
            pos += 4
        else:
            ns = nr = 0
        if data[0] & 0x04:
            off, = struct.unpack('>H', data[pos:pos+2])
            pos += 2+off
        return tunnel_id, session_id, ns, nr, data[pos:]

def make_l2tp(tunnel_id, session_id, ns, nr, body):
    if type(body) is dict:
        data = bytearray()
        for k, v in sorted(body.items()):
            if k in (0, 2, 9, 10, 14):
                v = struct.pack('>H', v)
            elif k in (3, 15, 19, 24):
                v = struct.pack('>I', v)
            data.extend(struct.pack('>HHH', (len(v)+6)|0x8000, 0, k) + v)
        header = struct.pack('>HHHHHH', 0xc802, len(data)+12, tunnel_id, session_id, ns, nr)
    else:
        data = body
        header = struct.pack('>HHHH', 0x4002, len(data)+8, tunnel_id, session_id)
    return header + data

class State(enum.Enum):
    LISTEN = 0
    SYN_SENT = 1
    SYN_RECEIVED = 2
    ESTABLISHED = 3
    FIN_WAIT_1 = 4
    FIN_WAIT_2 = 5
    CLOSE_WAIT = 6
    CLOSING = 7
    LAST_ACK = 8
    TIME_WAIT = 9
    CLOSED = 10
    INITIAL = 100

class Control(enum.IntFlag):
    FIN = 0x01
    SYN = 0x02
    RST = 0x04
    PSH = 0x08
    ACK = 0x10
    URG = 0x20

class TCPStack:
    def __init__(self, src_ip, src_port, dst_ip, dst_name, dst_port, reply, tcp_conn, verbose):
        self.src_ip = src_ip
        self.src_port = src_port
        self.dst_ip = dst_ip
        self.dst_name = dst_name
        self.dst_port = dst_port
        self.reply = reply
        self.tcp_conn = tcp_conn
        self.src_seq = 0
        self.dst_seq = self.dst_ack = 0
        self.src_win = []
        self.dst_win = collections.deque()
        self.dst_win_buf = bytearray()
        self.writer = None
        self.state = State.INITIAL
        self.cwnd = 2*SMSS if SMSS>2190 else 3*SMSS if SMSS>1095 else 4*SMSS
        self.rwnd = 65535
        self.ssthresh = 65535
        self.fast_resend = 0
        self.wait_send = asyncio.Event()
        self.wait_ack = asyncio.Event()
        self.wait_fast = asyncio.Event()
        self.rto = 3
        self.srtt = self.rttvar = None
        self.update = time.perf_counter()
        self.verbose = verbose
    def logwrite(self, data):
        if self.verbose:
            print(f'TCP WRITE {self.dst_name}:{self.dst_port} {data}')
    def logread(self, data):
        if self.verbose:
            print(f'TCP READ {self.dst_name}:{self.dst_port} {data}')
    def obsolete(self):
        return self.state == State.CLOSED or time.perf_counter() - self.update > 600
    def close(self):
        self.state = State.CLOSED
    def calc_rto(self, r):
        if self.srtt is None:
            self.srtt = r
            self.rttvar = r/2
        else:
            self.rttvar = 0.75*self.rttvar + 0.25*abs(self.srtt-r)
            self.srtt = 0.875*self.srtt + 0.125*r
        self.rto = min(self.srtt + max(0.1, 4*self.rttvar), 60)
    async def retransmit(self):
        while self.state != State.CLOSED:
            if self.dst_seq == self.dst_ack:
                self.wait_ack.clear()
                await self.wait_ack.wait()
            if self.state == State.CLOSED:
                break
            seq = self.dst_ack
            self.wait_fast.clear()
            timeout = False
            try:
                await asyncio.wait_for(self.wait_fast.wait(), self.rto)
            except asyncio.TimeoutError:
                timeout = True
            if self.state == State.CLOSED:
                break
            if seq != self.dst_ack:
                continue
            # retransmit
            #print('retransmit', self.dst_ip, self.dst_port, self.dst_ack, self.dst_seq, len(self.dst_win_buf), self.rto, timeout)
            flag = None
            if self.dst_win:
                self.dst_win[0][3] = 0
                if self.dst_win[0][0] == self.dst_ack and self.dst_win[0][2] == 1:
                    flag = self.dst_win_buf[0]
            if flag is not None:
                self.send(seq=self.dst_ack, flag=flag)
            else:
                tcp_body = self.dst_win_buf[:SMSS]
                self.send(tcp_body, seq=self.dst_ack)
            if timeout:
                self.rto = min(self.rto*2, 60)
                self.ssthresh = max((self.dst_seq-self.dst_ack)//2, 2*SMSS)
                self.cwnd = self.ssthresh+3*SMSS
    def parse(self, ip_body):
        src_port, dst_port, seq, ack, offset, flag, window = struct.unpack('>HHIIBBH', ip_body[:16])
        tcp_body = ip_body[offset>>2:]
        self.rwnd = window
        self.update = time.perf_counter()
        # print('RECV', self.dst_name, self.dst_port, self.state, Control(flag), seq, ack, len(tcp_body))
        if self.state == State.CLOSED:
            if flag & Control.RST:
                pass
            elif flag & Control.ACK:
                self.send(seq=ack, flag=Control.RST)
            else:
                self.send(seq=0, ack=seq+len(tcp_body), flag=Control.RST|Control.ACK)
        elif self.state == State.INITIAL:
            if flag & Control.RST:
                pass
            elif flag & Control.ACK:
                self.send(seq=ack, flag=Control.RST)
            elif flag & Control.SYN:
                self.state = State.SYN_RECEIVED
                self.src_seq = seq+1
                self.dst_seq = self.dst_ack = random.randrange(0x100000000)
                asyncio.ensure_future(self.connect())
                asyncio.ensure_future(self.retransmit())
        elif flag & Control.RST:
            self.close()
            self.wait_ack.set()
            self.wait_send.set()
            try:
                self.writer.close()
            except Exception:
                pass
        elif flag & Control.SYN:
            pass
        elif flag & Control.ACK == 0:
            pass
        else:
            if self.state == State.SYN_RECEIVED:
                self.state = State.ESTABLISHED
            while self.dst_ack-ack > 0x20000000:
                ack += 0x100000000
            if self.dst_ack < self.dst_seq:
                diff_ack = ack-self.dst_ack
                if diff_ack == 0 and not tcp_body:
                    self.fast_resend += 1
                    if self.fast_resend < 3:
                        self.wait_send.set()
                    elif self.fast_resend == 3:
                        self.wait_fast.set()
                    else:
                        self.cwnd += SMSS
                if diff_ack > 0:
                    self.fast_resend = 0
                    counter = 0
                    while self.dst_win and self.dst_win[0][0]+self.dst_win[0][1] <= ack:
                        seq, length, tp, counter = self.dst_win.popleft()
                    del self.dst_win_buf[:diff_ack]
                    self.dst_ack = ack
                    self.wait_fast.set()
                    if self.dst_seq-self.dst_ack <= min(self.cwnd, self.rwnd):
                        self.wait_send.set()
                    if self.cwnd < self.ssthresh:
                        self.cwnd += min(diff_ack, SMSS)
                    else:
                        self.cwnd += SMSS*SMSS//self.cwnd
                    if counter == 0:
                        self.cwnd = self.ssthresh
                    if counter > 0:
                        time_diff = time.perf_counter() - counter
                        self.calc_rto(time_diff)
            if self.state == State.FIN_WAIT_1:
                if self.dst_ack >= self.dst_seq:
                    self.state = State.FIN_WAIT_2
            if tcp_body and self.state in (State.ESTABLISHED, State.FIN_WAIT_1, State.FIN_WAIT_2):
                while self.src_seq-seq > 0x20000000:
                    seq += 0x100000000
                if seq+len(tcp_body) <= self.src_seq:
                    pass
                elif seq <= self.src_seq:
                    self.logwrite(tcp_body[self.src_seq-seq:])
                    self.writer.write(tcp_body[self.src_seq-seq:])
                    self.src_seq = seq+len(tcp_body)
                    while self.src_win and self.src_win[0][0] <= self.src_seq:
                        seq, tcp_body = heapq.heappop(self.src_win)
                        if seq+len(tcp_body) > self.src_seq:
                            self.logwrite(tcp_body[self.src_seq-seq:])
                            self.writer.write(tcp_body[self.src_seq-seq:])
                            self.src_seq = seq+len(tcp_body)
                elif seq-self.src_seq < 0x20000000:
                    heapq.heappush(self.src_win, (seq, tcp_body))
                self.send()
            if flag & Control.FIN:
                while self.src_seq-seq > 0x20000000:
                    seq += 0x100000000
                if seq+1 <= self.src_seq:
                    pass
                elif seq <= self.src_seq:
                    self.src_seq = seq+1
                    self.send()
                    if self.state in (State.SYN_RECEIVED, State.ESTABLISHED):
                        self.state = State.CLOSE_WAIT
                        try:
                            self.writer.close()
                        except Exception:
                            pass
                    elif self.state == State.FIN_WAIT_2:
                        self.close()
                        self.wait_ack.set()
    def send(self, tcp_body=b'', *, flag=Control.ACK, seq=None, ack=None):
        self.update = time.perf_counter()
        window = max(0, (65535-len(self.writer.transport._buffer)) if self.writer else 0)
        # print('SEND', self.dst_name, self.dst_port, self.state, Control(flag), (self.dst_seq if seq is None else seq), (self.src_seq if ack is None else ack), len(tcp_body))
        tcp_header = struct.pack('>HHIIBBHHH', self.dst_port, self.src_port, (self.dst_seq if seq is None else seq)&0xffffffff, (self.src_seq if ack is None else ack)&0xffffffff, 5<<4, flag, window, 0, 0)
        ip_body = bytearray(tcp_header + tcp_body)
        tochecksum = bytearray(self.dst_ip.packed+self.src_ip.packed+b'\x00\x06'+len(ip_body).to_bytes(2, 'big') + ip_body)
        if len(tochecksum) % 2 == 1:
            tochecksum.extend(b'\x00')
        ip_body[16:18] = checksum(tochecksum)
        data = make_ipv4(6, self.dst_ip, self.src_ip, ip_body)
        if not self.reply(data):
            self.close()
            self.wait_ack.set()
            self.wait_send.set()
            try:
                self.writer.close()
            except Exception:
                pass
    async def connect(self):
        # print(f'connect {self.dst_ip}:{self.dst_port}')
        total = 0
        try:
            reader, self.writer = await self.tcp_conn.tcp_connect(self.dst_name, self.dst_port)
        except Exception:
            # connect fail
            self.close()
            self.wait_ack.set()
            self.send(flag=Control.RST)
            return
        self.send(flag=Control.ACK|Control.SYN)
        self.dst_win.append([self.dst_seq, 1, 1, time.perf_counter()])
        self.dst_win_buf.extend(bytes([Control.ACK|Control.SYN]))
        self.dst_seq += 1
        self.wait_ack.set()
        while True:
            try:
                data = await reader.read_()
            except Exception:
                data = None
            if not data:
                break
            self.logread(data)
            # print(f'TCP READ {self.dst_name}:{self.dst_port} {data}')
            data = bytearray(data)
            while data:
                if self.dst_seq-self.dst_ack > min(self.cwnd, self.rwnd):
                    self.wait_send.clear()
                    await self.wait_send.wait()
                    if self.state == State.CLOSED:
                        break
                tcp_body = data[:SMSS]
                del data[:SMSS]
                total += len(tcp_body)
                self.send(tcp_body)
                self.dst_win.append([self.dst_seq, len(tcp_body), 0, time.perf_counter()])
                self.dst_win_buf.extend(tcp_body)
                self.dst_seq += len(tcp_body)
                self.wait_ack.set()
        if self.state == State.CLOSE_WAIT:
            self.send(flag=Control.ACK|Control.FIN)
            self.dst_win_buf.extend(bytes([Control.ACK|Control.FIN]))
            self.dst_seq += 1
            self.wait_ack.set()
            self.close()
        elif self.state != State.CLOSED:
            self.send(flag=Control.ACK|Control.FIN)
            self.state = State.FIN_WAIT_1
            self.dst_win_buf.extend(bytes([Control.ACK|Control.FIN]))
            self.dst_seq += 1

class IPPacket:
    def __init__(self, args):
        self.tcp_stack = {}
        self.dns_server = args.dns
        self.dns_cache = None if args.nocache else dns.DNSCache()
        self.salgorithm = args.salgorithm
        self.rserver = args.rserver
        self.urserver = args.urserver
        self.DIRECT = args.DIRECT
        self.verbose = bool(args.v)
    def schedule(self, host_name, port, udp=False):
        rserver = self.urserver if udp else self.rserver
        filter_cond = lambda o: o.alive and (not o.match or o.match(host_name) or o.match(str(port)))
        if self.salgorithm == 'fa':
            return next(filter(filter_cond, rserver), None)
        elif self.salgorithm == 'rr':
            for i, roption in enumerate(rserver):
                if filter_cond(roption):
                    rserver.append(rserver.pop(i))
                    return roption
        elif self.salgorithm == 'rc':
            filters = [i for i in rserver if filter_cond(i)]
            return random.choice(filters) if filters else None
        elif self.salgorithm == 'lc':
            return min(filter(filter_cond, rserver), default=None, key=lambda i: i.total)
        else:
            raise Exception('Unknown scheduling algorithm') #Unreachable
    def handle_ipv4(self, remote_id, data, reply):
        proto, src_ip, dst_ip, ip_body = parse_ipv4(data)
        dst_name = self.dns_cache.ip2domain(str(dst_ip)) if self.dns_cache else str(dst_ip)
        if proto == enums.IpProto.UDP:
            src_port, dst_port, udp_body = parse_udp(ip_body)
            option = self.schedule(dst_name, dst_port, udp=True) or self.DIRECT
            key = (remote_id[0], remote_id[1], src_port)
            if dst_port == 53:
                try:
                    record = dns.DNSRecord.unpack(udp_body)
                    answer = self.dns_cache.query(record) if self.dns_cache else None
                    if self.verbose:
                        print(f'DNS {remote_id[0]}:{src_port}{option.logtext(dst_name, dst_port)} Query={record.q.qname}{" (Cached)" if answer else ""}')
                    if answer:
                        ip_body = make_udp(dst_port, src_port, answer.pack())
                        data = make_ipv4(proto, dst_ip, src_ip, ip_body)
                        reply(data)
                        return
                except Exception as e:
                    print(e)
            else:
                if self.verbose:
                    print(f'UDP {remote_id[0]}:{src_port}{option.logtext(dst_name, dst_port)} Length={len(udp_body)}')
            def udp_reply(udp_body):
                if dst_port == 53:
                    record = dns.DNSRecord.unpack(udp_body)
                    self.dns_cache.answer(record) if self.dns_cache else None
                    if self.verbose:
                        print(f'DNS {remote_id[0]}:{src_port}{option.logtext(dst_name, dst_port).replace("->","<-")} Answer=['+' '.join(f'{r.rname}->{r.rdata}' for r in record.rr)+']')
                else:
                    if self.verbose:
                        print(f'UDP {remote_id[0]}:{src_port}{option.logtext(dst_name, dst_port).replace("->","<-")} Length={len(udp_body)}')
                ip_body = make_udp(dst_port, src_port, udp_body)
                data = make_ipv4(proto, dst_ip, src_ip, ip_body)
                reply(data)
            asyncio.ensure_future(option.udp_sendto(dst_name, dst_port, udp_body, udp_reply, key))
        elif proto == enums.IpProto.TCP:
            src_port, dst_port, flag, tcp_body = parse_tcp(ip_body)
            key = (remote_id[0], remote_id[1], src_port)
            tcp = self.tcp_stack.get(key)
            if tcp is None:
                if flag & Control.SYN == 0:
                    return
                option = self.schedule(dst_name, dst_port) or self.DIRECT
                print(f'TCP {remote_id[0]}:{src_port}{option.logtext(dst_name, dst_port)}')
                for spi, tcp in list(self.tcp_stack.items()):
                    if tcp.obsolete():
                        self.tcp_stack.pop(spi)
                self.tcp_stack[key] = tcp = TCPStack(src_ip, src_port, dst_ip, dst_name, dst_port, reply, option, self.verbose)
                #print(f'TCP Connections = {len(self.tcp_stack)}')
            tcp.parse(ip_body)
        elif proto == enums.IpProto.ICMP:
            icmptp, code, icmp_body = parse_icmp(ip_body)
            if icmptp == 0:
                tid, seq = struct.unpack('>HH', ip_body[4:8])
                if self.verbose:
                    print(f'PING {remote_id[0]} -> {dst_name} Id={tid} Seq={seq} Data={icmp_body}')
            elif icmptp == 8:
                tid, seq = struct.unpack('>HH', ip_body[4:8])
                if self.verbose:
                    print(f'ECHO {remote_id[0]} -> {dst_name} Id={tid} Seq={seq} Data={icmp_body}')
                # NEED ROOT PRIVILEGE TO SEND ICMP PACKET
                # a = socket.socket(socket.AF_INET, socket.SOCK_RAW, proto)
                # a.sendto(icmp_body, (dst_name, 1))
                # a.close()
            elif icmptp == 3 and code == 3:
                eproto, esrc_ip, edst_ip, eip_body = parse_ipv4(icmp_body)
                eport = int.from_bytes(eip_body[2:4], 'big')
                if self.verbose:
                    print(f'ICMP {remote_id[0]} -> {dst_name} {eproto.name} :{eport} Denied')
            else:
                if self.verbose:
                    print(f'ICMP {remote_id[0]} -> {dst_name} Data={ip_body}')
        else:
            print(f'{enums.IpProto(proto).name} -> {dst_name} Data={data}')
    def handle_l2tp(self, remote_id, data, reply):
        src_port, dst_port, udp_body = parse_udp(data)
        tunnel_id, session_id, ns, nr, l2tp_body = parse_l2tp(udp_body)
        # print(tunnel_id, session_id, ns, nr, l2tp_body)
        def reply_l2tp(l2tp_body):
            # print('reply', l2tp_body)
            ns_nr = type(l2tp_body) is dict
            udp_body = make_l2tp(tunnel_id, session_id, nr if ns_nr else None, ns+1 if ns_nr else None, l2tp_body)
            ip_body = make_udp(dst_port, src_port, udp_body)
            return reply(ip_body)
        if type(l2tp_body) is dict:
            msgtp = l2tp_body[enums.L2TPAttr.MsgType]
            if msgtp == enums.L2TPType.SCCRQ:
                tunnel_id = l2tp_body[enums.L2TPAttr.TunnelID]
                l2tp_body[enums.L2TPAttr.MsgType] = enums.L2TPType.SCCRP
                l2tp_body[enums.L2TPAttr.HostName] = b'python-esp\x00'
                reply_l2tp(l2tp_body)
            elif msgtp in (enums.L2TPType.SCCCN, enums.L2TPType.ICCN):
                reply_l2tp({})
            elif msgtp == enums.L2TPType.ICRQ:
                session_id = l2tp_body[enums.L2TPAttr.SessionID]
                l2tp_body[enums.L2TPAttr.MsgType] = enums.L2TPType.ICRP
                l2tp_body.pop(enums.L2TPAttr.CallSerial)
                reply_l2tp(l2tp_body)
        else:
            HEAD = struct.Struct('>BBH')
            def parse_lcp(body, reply_body):
                code, mid, mlen = HEAD.unpack(body[:4])
                magic = b'PESP'
                if code == 1:
                    l2tp_body = HEAD.pack(2, mid, len(body)) + body[4:]
                    reply_body(l2tp_body)
                    s = b'\x02\x06\x00\x00\x00\x00\x05\x06'+magic+b'\x07\x02\x08\x02'
                    reply_body(HEAD.pack(1, mid+1, len(s)+4) + s)
                elif code == 2:
                    reply_body(HEAD.pack(9, mid, 8) + magic)
                elif code == 9:
                    reply_body(HEAD.pack(10, mid, 8) + magic)
            def parse_ccp(body, reply_body):
                code, mid, mlen = HEAD.unpack(body[:4])
                if code == 1:
                    reply_body(HEAD.pack(2, mid, len(body)) + body[4:])
                    reply_body(HEAD.pack(1, mid+1, 4))
            def parse_ipcp(body, reply_body):
                code, mid, mlen = HEAD.unpack(body[:4])
                if code == 1:
                    if b'\x03\x06\x00\x00\x00\x00' not in body:
                        reply_body(HEAD.pack(2, mid, len(body)) + body[4:])
                    else:
                        addr = ipaddress.ip_address('10.0.0.1').packed
                        dns = ipaddress.ip_address(self.dns_server).packed
                        s = b'\x03\x06'+addr+b'\x81\x06'+dns+b'\x83\x06\x00\x00\x00\x00'
                        reply_body(HEAD.pack(3, mid, len(s)+4) + s)
                        addr = ipaddress.ip_address('10.0.0.2').packed
                        reply_body(HEAD.pack(1, mid+1, 10) + b'\x03\x06' + addr)
            def parse_ipv6cp(body, reply_body):
                code, mid, mlen = HEAD.unpack(body[:4])
                if code == 1:
                    reply_body(HEAD.pack(2, mid, len(body)) + body[4:])
                    reply_body(HEAD.pack(1, mid, 14) + b'\x01\x0a' + os.urandom(8))
            def parse_ip(body, reply_body):
                self.handle_ipv4(remote_id, body, reply_body)
            CONTROLS = (( b'\xff\x03\xc0\x21', parse_lcp ),
                        ( b'\xff\x03\x80\xfd', parse_ccp ),
                        ( b'\x80\x21', parse_ipcp ),
                        ( b'\xff\x03\x80\x21', parse_ipcp ),
                        ( b'\x80\x57', parse_ipv6cp ),
                        ( b'\x21', parse_ip),
                        ( b'\xff\x03\x00\x21', parse_ip),
                       )
            for head, control in CONTROLS:
                if l2tp_body.startswith(head):
                    body = l2tp_body[len(head):]
                    control(body, lambda data: reply_l2tp(head+data))
    def handle(self, remote_id, header, data, reply):
        if header == enums.IpProto.IPV4:
            self.handle_ipv4(remote_id, data, reply)
        elif header == enums.IpProto.UDP:
            src_port, dst_port, udp_body = parse_udp(data)
            if dst_port == 1701:
                self.handle_l2tp(remote_id, data, reply)
            else:
                print(f'UDP Unhandled Port={dst_port}. Data={udp_body}')
        else:
            print(f'{enums.IpProto(header).name} Unhandled Protocol. Data={data}')



import struct, io, collections, os, random, ipaddress
from . import enums

class Payload:
    def __init__(self, type, critical=False):
        self.type = enums.Payload(type)
        self.critical = critical
        self.data = None
    @classmethod
    def parse(cls, type, critical, stream, length):
        self = cls.__new__(cls)
        Payload.__init__(self, type, critical)
        self.parse_data(stream, length)
        return self
    def parse_data(self, stream, length):
        self.data = stream.read(length)
    def to_bytes(self):
        return self.data
    def to_repr(self):
        return f'data={self.data}'
    def __repr__(self):
        return f'{self.type.name}({"critical, " if self.critical else ""}{self.to_repr()})'

def attr_parse(stream, length, attr_type_cls):
    values = collections.OrderedDict()
    while length > 0:
        attr_type, value = struct.unpack('>HH', stream.read(4))
        length -= 4
        if attr_type & 0x8000:
            attr_type &= 0x7FFF
        else:
            length -= value
            value = stream.read(value)
        values[attr_type_cls(attr_type)] = value
    return values

def attr_to_bytes(values):
    return b''.join((struct.pack('>HH', i|0x8000, j) if isinstance(j, int) else struct.pack('>HH', i, len(j))+j) for i, j in values.items())

Transform_1 = collections.namedtuple('Transform', 'num id values')

class Proposal_1:
    def __init__(self, num, protocol, spi, transforms):
        self.num = num
        self.protocol = enums.Protocol(protocol)
        self.spi = spi
        self.transforms = transforms
    @classmethod
    def parse(cls, stream):
        num, protocol, spi_size, n_transforms = struct.unpack('>BBBB', stream.read(4))
        spi = stream.read(spi_size)
        transforms = []
        more = True
        while more:
            more, length, tnum, id = struct.unpack('>BxHBB2x', stream.read(8))
            if protocol == enums.Protocol.ESP:
                values = attr_parse(stream, length-8, enums.ESPAttr)
                for attr_type in values:
                    if attr_type in enums.ESPTable_1:
                        values[attr_type] = enums.ESPTable_1[attr_type](values[attr_type])
                transforms.append(Transform_1(tnum, enums.EncrId(id), values))
            else:
                values = attr_parse(stream, length-8, enums.TransformAttr)
                for attr_type in values:
                    if attr_type in enums.TransformTable_1:
                        values[attr_type] = enums.TransformTable_1[attr_type](values[attr_type])
                transforms.append(Transform_1(tnum, enums.Protocol(id), values))
        return Proposal_1(num, protocol, spi, transforms)
    def to_bytes(self):
        data = bytearray()
        data.extend(struct.pack('>BBBB', self.num, self.protocol, len(self.spi), len(self.transforms)))
        data.extend(self.spi)
        for idx, transform in enumerate(self.transforms):
            transform_data = attr_to_bytes(transform.values)
            data.extend(struct.pack('>BxHBB2x', 0 if idx==len(self.transforms)-1 else 3,
                len(transform_data)+8, transform.num, transform.id))
            data.extend(transform_data)
        return data
    def to_repr(self):
        return f'{self.protocol.name}:{self.num}(spi={self.spi.hex() or "None"}, ' + ', '.join(
            f'{i.id.name}:{i.num}({", ".join(j.name+"="+str(k) for j,k in i.values.items())})' for i in self.transforms) + ')'

class PayloadSA_1(Payload):
    def __init__(self, doi, situation, proposals):
        Payload.__init__(self, enums.Payload.SA_1)
        self.doi = doi
        self.situation = situation
        self.proposals = proposals
    def parse_data(self, stream, length):
        self.doi, self.situation = struct.unpack('>II', stream.read(8))
        self.proposals = []
        more = True
        while more:
            more, length = struct.unpack('>BxH', stream.read(4))
            self.proposals.append(Proposal_1.parse(stream))
    def to_bytes(self):
        data = bytearray(struct.pack('>II', self.doi, self.situation))
        for idx, proposal in enumerate(self.proposals):
            proposal_data = proposal.to_bytes()
            data.extend(struct.pack('>BxH', 0 if idx==len(self.proposals)-1 else 2, len(proposal_data)+4))
            data.extend(proposal_data)
        return data
    def to_repr(self):
        return f'doi={self.doi}, situation={self.situation}, ' + ', '.join(i.to_repr() for i in self.proposals)

class PayloadKE_1(Payload):
    def __init__(self, ke_data):
        Payload.__init__(self, enums.Payload.KE_1)
        self.ke_data = ke_data
    def parse_data(self, stream, length):
        self.ke_data = stream.read(length)
    def to_bytes(self):
        return self.ke_data
    def to_repr(self):
        return f'{self.ke_data.hex()}'

class PayloadID_1(Payload):
    def __init__(self, id_type, id_data, prot=0, port=0, critical=False):
        Payload.__init__(self, enums.Payload.ID_1, critical)
        self.id_type = enums.IDType(id_type)
        self.prot = enums.IpProto(prot)
        self.port = port
        self.id_data = id_data
    def parse_data(self, stream, length):
        id_type, prot, self.port = struct.unpack('>BBH', stream.read(4))
        self.id_type = enums.IDType(id_type)
        self.prot = enums.IpProto(prot)
        self.id_data = stream.read(length-4)
    def to_bytes(self):
        return struct.pack('>BBH', self.id_type, self.prot, self.port) + self.id_data
    def _id_data_str(self):
        if self.id_type in (enums.IDType.ID_RFC822_ADDR, enums.IDType.ID_FQDN):
            return self.id_data.decode()
        elif self.id_type in (enums.IDType.ID_IPV4_ADDR, enums.IDType.ID_IPV6_ADDR):
            return str(ipaddress.ip_address(self.id_data))+(f':{self.port}({self.prot.name})' if self.prot!=0 else '')
        else:
            return self.id_data.hex()
    def to_repr(self):
        return f'{self.id_type.name}({self._id_data_str()})'

class PayloadHASH_1(Payload):
    def __init__(self, data):
        Payload.__init__(self, enums.Payload.HASH_1)
        self.data = data
    def parse_data(self, stream, length):
        self.data = stream.read(length)
    def to_bytes(self):
        return self.data
    def to_repr(self):
        return f'{self.data.hex()}'

class PayloadNONCE_1(Payload):
    def __init__(self, nonce=None, critical=False):
        Payload.__init__(self, enums.Payload.NONCE_1, critical)
        self.nonce = os.urandom(random.randrange(16, 256)) if nonce is None else nonce
    def parse_data(self, stream, length):
        self.nonce = stream.read(length)
    def to_bytes(self):
        return self.nonce
    def to_repr(self):
        return f'{self.nonce.hex()}'

class PayloadNOTIFY_1(Payload):
    def __init__(self, doi, protocol, notify, spi, data):
        Payload.__init__(self, enums.Payload.NOTIFY_1)
        self.doi = doi
        self.protocol = enums.Protocol(protocol)
        self.notify = enums.Notify(notify)
        self.spi = spi
        self.data = data
    def parse_data(self, stream, length):
        self.doi, protocol, spi_size, notify = struct.unpack('>IBBH', stream.read(8))
        self.protocol = enums.Protocol(protocol)
        self.notify = enums.Notify(notify)
        self.spi = stream.read(spi_size)
        self.data = stream.read(length-8-spi_size)
    def to_bytes(self):
        data = bytearray(struct.pack('>IBBH', self.doi, self.protocol, len(self.spi), self.notify))
        data.extend(self.spi)
        data.extend(self.data)
        return data
    def to_repr(self):
        return f'{self.notify.name}(doi={self.doi}, {"protocol="+self.protocol.name+", " if self.protocol else ""}{"spi="+self.spi.hex()+", " if self.spi else ""}{"data="+self.data.hex() if self.data else ""})'

class PayloadDELETE_1(Payload):
    def __init__(self, doi, protocol, spis):
        Payload.__init__(self, enums.Payload.DELETE_1)
        self.doi = doi
        self.protocol = enums.Protocol(protocol)
        self.spis = spis
    def parse_data(self, stream, length):
        self.doi, protocol, spi_size, num_spis = struct.unpack('>IBBH', stream.read(8))
        self.protocol = enums.Protocol(protocol)
        self.spis = [stream.read(spi_size) for i in range(num_spis)]
    def to_bytes(self):
        data = bytearray()
        data.extend(struct.pack('>IBBH', self.doi, self.protocol, len(self.spis[0]) if self.spis else 0, len(self.spis)))
        for spi in self.spis:
            data.extend(spi)
        return data
    def to_repr(self):
        return f'{self.protocol.name}(doi={self.doi}, {", ".join(i.hex() for i in self.spis)})'

class PayloadVENDOR_1(Payload):
    def __init__(self, vendor):
        Payload.__init__(self, enums.Payload.VENDOR_1, False)
        self.vendor = vendor
    def parse_data(self, stream, length):
        self.vendor = stream.read(length)
    def to_bytes(self):
        return self.vendor
    def to_repr(self):
        return f'{self.vendor.hex()}'

class PayloadCP_1(Payload):
    def __init__(self, type, attrs, critical=False, identifier=0):
        Payload.__init__(self, enums.Payload.CP_1, critical)
        self.cftype = enums.CFGType(type)
        self.identifier = identifier
        self.attrs = attrs
    def parse_data(self, stream, length):
        cftype, self.identifier = struct.unpack('>BxH', stream.read(4))
        self.cftype = enums.CFGType(cftype)
        self.attrs = attr_parse(stream, length-4, enums.CPAttrType)
    def to_bytes(self):
        return struct.pack('>BxH', self.cftype, self.identifier) + attr_to_bytes(self.attrs)
    def to_repr(self):
        return f'{self.cftype.name}(id={self.identifier}, {", ".join(k.name+"="+(str(v) if type(v) is int else (v.hex() or "None")) for k, v in self.attrs.items())})'

class PayloadNATD_1(Payload):
    def __init__(self, data):
        Payload.__init__(self, enums.Payload.NATD_1, False)
        self.data = data
    def parse_data(self, stream, length):
        self.data = stream.read(length)
    def to_bytes(self):
        return self.data
    def to_repr(self):
        return f'{self.data.hex()}'


Transform = collections.namedtuple('Transform', 'type id keylen')

class Proposal:
    def __init__(self, num, protocol, spi, transforms):
        self.num = num
        self.protocol = enums.Protocol(protocol)
        self.spi = spi
        self.transforms = transforms
    @classmethod
    def parse(cls, stream):
        num, protocol, spi_size, n_transforms = struct.unpack('>BBBB', stream.read(4))
        spi = stream.read(spi_size)
        transforms = []
        more = True
        while more:
            more, length, type, id = struct.unpack('>BxHBxH', stream.read(8))
            values = attr_parse(stream, length-8, enums.TransformAttr)
            keylen = values.get(enums.TransformAttr.KEY_LENGTH)
            transforms.append(Transform(enums.Transform(type), enums.TransformTable[type](id), keylen))
        return Proposal(num, protocol, spi, transforms)
    def to_bytes(self):
        data = bytearray()
        data.extend(struct.pack('>BBBB', self.num, self.protocol, len(self.spi), len(self.transforms)))
        data.extend(self.spi)
        for idx, transform in enumerate(self.transforms):
            transform_data = struct.pack('>HH', 0x800e, transform.keylen) if transform.keylen else b''
            data.extend(struct.pack('>BxHBxH', 0 if idx==len(self.transforms)-1 else 3,
                len(transform_data)+8, transform.type, transform.id))
            data.extend(transform_data)
        return data
    def to_repr(self):
        return f'{self.protocol.name}:{self.num}(spi={self.spi.hex() or "None"}, ' + ', '.join(
            f'{i.id.name}{"(keylen="+str(i.keylen)+")" if i.keylen else ""}' for i in self.transforms) + ')'
    def remove_redundancy(self):
        transforms = []
        transformtypes = set()
        for t in self.transforms:
            if t.type not in transformtypes:
                transformtypes.add(t.type)
                transforms.append(t)
        return Proposal(self.num, self.protocol, self.spi, transforms)
    def get_transform(self, type):
        return next((x for x in self.transforms if x.type == type), None)

class PayloadSA(Payload):
    def __init__(self, proposals, critical=False):
        Payload.__init__(self, enums.Payload.SA, critical)
        self.proposals = proposals
    def parse_data(self, stream, length):
        self.proposals = []
        more = True
        while more:
            more, length = struct.unpack('>BxH', stream.read(4))
            self.proposals.append(Proposal.parse(stream))
    def get_proposal(self, encr_id):
        for i in self.proposals:
            if i.get_transform(enums.Transform.ENCR).id == encr_id:
                return i.remove_redundancy()
    def to_bytes(self):
        data = bytearray()
        for idx, proposal in enumerate(self.proposals):
            proposal_data = proposal.to_bytes()
            data.extend(struct.pack('>BxH', 0 if idx==len(self.proposals)-1 else 2, len(proposal_data)+4))
            data.extend(proposal_data)
        return data
    def to_repr(self):
        return ', '.join(i.to_repr() for i in self.proposals)

class PayloadKE(Payload):
    def __init__(self, dh_group, ke_data, critical=False):
        Payload.__init__(self, enums.Payload.KE, critical)
        self.dh_group = dh_group
        self.ke_data = ke_data
    def parse_data(self, stream, length):
        self.dh_group, = struct.unpack('>H2x', stream.read(4))
        self.ke_data = stream.read(length-4)
    def to_bytes(self):
        return struct.pack('>H2x', self.dh_group) + self.ke_data
    def to_repr(self):
        return f'{self.dh_group}, {self.ke_data.hex()}'

class PayloadIDi(PayloadID_1):
    def __init__(self, id_type, id_data, prot=0, port=0, critical=False):
        PayloadID_1.__init__(self, id_type, id_data, prot, port, critical)
        self.type = enums.Payload.IDi

class PayloadIDr(PayloadID_1):
    def __init__(self, id_type, id_data, prot=0, port=0, critical=False):
        PayloadID_1.__init__(self, id_type, id_data, prot, port, critical)
        self.type = enums.Payload.IDr

class PayloadAUTH(Payload):
    def __init__(self, method, auth_data, critical=False):
        Payload.__init__(self, enums.Payload.AUTH, critical)
        self.method = enums.AuthMethod(method)
        self.auth_data = auth_data
    def parse_data(self, stream, length):
        self.method = enums.AuthMethod(struct.unpack('>B3x', stream.read(4))[0])
        self.auth_data = stream.read(length-4)
    def to_bytes(self):
        return struct.pack('>B3x', self.method) + self.auth_data
    def to_repr(self):
        return f'{self.method.name}({self.auth_data.hex()})'

class PayloadNONCE(PayloadNONCE_1):
    def __init__(self, nonce=None, critical=False):
        PayloadNONCE_1.__init__(self, nonce, critical)
        self.type = enums.Payload.NONCE

class PayloadNOTIFY(Payload):
    def __init__(self, protocol, notify, spi, data, critical=False):
        Payload.__init__(self, enums.Payload.NOTIFY, critical)
        self.protocol = enums.Protocol(protocol)
        self.notify = enums.Notify(notify)
        self.spi = spi
        self.data = data
    def parse_data(self, stream, length):
        protocol, spi_size, notify = struct.unpack('>BBH', stream.read(4))
        self.protocol = enums.Protocol(protocol)
        self.notify = enums.Notify(notify)
        self.spi = stream.read(spi_size)
        self.data = stream.read(length-4-spi_size)
    def to_bytes(self):
        data = bytearray(struct.pack('>BBH', self.protocol, len(self.spi), self.notify))
        data.extend(self.spi)
        data.extend(self.data)
        return data
    def to_repr(self):
        return f'{self.notify.name}({"protocol="+self.protocol.name+", " if self.protocol else ""}{"spi="+self.spi.hex()+", " if self.spi else ""}{"data="+self.data.hex() if self.data else ""})'

class PayloadDELETE(Payload):
    def __init__(self, protocol, spis, critical=False):
        Payload.__init__(self, enums.Payload.DELETE, critical)
        self.protocol = enums.Protocol(protocol)
        self.spis = spis
    def parse_data(self, stream, length):
        protocol, spi_size, num_spis = struct.unpack('>BBH', stream.read(4))
        self.protocol = enums.Protocol(protocol)
        self.spis = [stream.read(spi_size) for i in range(num_spis)]
    def to_bytes(self):
        data = bytearray()
        data.extend(struct.pack('>BBH', self.protocol, len(self.spis[0]) if self.spis else 0, len(self.spis)))
        for spi in self.spis:
            data.extend(spi)
        return data
    def to_repr(self):
        return f'{self.protocol.name}({", ".join(i.hex() for i in self.spis)})'

class PayloadVENDOR(Payload):
    def __init__(self, vendor=None, critical=False):
        Payload.__init__(self, enums.Payload.VENDOR, critical)
        self.vendor = vendor
    def parse_data(self, stream, length):
        self.vendor = stream.read(length)
    def to_bytes(self):
        return self.vendor
    def to_repr(self):
        return f'{self.vendor.decode()}'

class TrafficSelector:
    def __init__(self, ts_type, ip_proto, start_port, end_port, start_addr, end_addr):
        self.ts_type = enums.TSType(ts_type)
        self.ip_proto = enums.IpProto(ip_proto)
        self.start_port = start_port
        self.end_port = end_port
        self.start_addr = start_addr
        self.end_addr = end_addr
    @classmethod
    def from_network(cls, subnet, port, ip_proto):
        return TrafficSelector(enums.TSType.TS_IPV4_ADDR_RANGE, ip_proto, port,
                               65535 if port == 0 else port, subnet[0], subnet[-1])
    def get_network(self):
        network = ipaddress.ip_network(self.start_addr)
        while self.end_addr not in network:
            network = network.supernet()
        return network
    def get_port(self):
        return 0 if self.start_port==0 and self.end_port==65535 else self.end_port
    @classmethod
    def parse(cls, stream):
        ts_type, ip_proto, length, start_port, end_port = struct.unpack('>BBHHH', stream.read(8))
        addr_len = (length - 8) // 2
        start_addr = stream.read(addr_len)
        end_addr = stream.read(addr_len)
        return TrafficSelector(ts_type, ip_proto, start_port, end_port, ipaddress.ip_address(start_addr),
                               ipaddress.ip_address(end_addr))
    def to_bytes(self):
        pack_addr = self.start_addr.packed + self.end_addr.packed
        return struct.pack('>BBHHH', self.ts_type, self.ip_proto, 8 + len(pack_addr), self.start_port, self.end_port) + pack_addr
    def __repr__(self):
        return f'{self.ts_type.name}({self.ip_proto.name}, {self.start_addr} - {self.end_addr}, {self.start_port} - {self.end_port})'

class PayloadTSi(Payload):
    def __init__(self, traffic_selectors, critical=False):
        Payload.__init__(self, enums.Payload.TSi, critical)
        self.traffic_selectors = traffic_selectors
    def parse_data(self, stream, length):
        n_ts, = struct.unpack_from('>B3x', stream.read(4))
        self.traffic_selectors = [TrafficSelector.parse(stream) for i in range(n_ts)]
    def to_bytes(self):
        data = bytearray(struct.pack('>BBH', len(self.traffic_selectors), 0, 0))
        for ts in self.traffic_selectors:
            data.extend(ts.to_bytes())
        return data
    def to_repr(self):
        return ', '.join(repr(i) for i in self.traffic_selectors)

class PayloadTSr(PayloadTSi):
    def __init__(self, traffic_selectors, critical=False):
        Payload.__init__(self, enums.Payload.TSr, critical)
        self.traffic_selectors = traffic_selectors

class PayloadSK(Payload):
    def __init__(self, ciphertext, critical=False):
        Payload.__init__(self, enums.Payload.SK, critical)
        self.ciphertext = ciphertext
    def parse_data(self, stream, length):
        self.ciphertext = stream.read(length)
    def to_bytes(self):
        return self.ciphertext
    def to_repr(self):
        return self.ciphertext.hex()

class PayloadCP(PayloadCP_1):
    def __init__(self, type, attrs, critical=False):
        PayloadCP_1.__init__(self, type, attrs, critical)
        self.type = enums.Payload.CP

class PayloadEAP(Payload):
    def __init__(self, code, data, critical=False):
        Payload.__init__(self, enums.Payload.EAP, critical)
        self.code = enums.EAPCode(code)
        self.data = data
    def parse_data(self, stream, length):
        self.code = enums.EAPCode(stream.unpack('>B3x', stream.read(4))[0])
        self.data = stream.read(length-4)
    def to_bytes(self):
        data = struct.pack('>BxH', self.code, len(self.data)+4)
        return data+self.data
    def to_repr(self):
        return f'{self.code.name}({self.data.hex()})'

PayloadClass = {
    enums.Payload.SA_1: PayloadSA_1,
    enums.Payload.KE_1: PayloadKE_1,
    enums.Payload.ID_1: PayloadID_1,
    enums.Payload.HASH_1: PayloadHASH_1,
    enums.Payload.NONCE_1: PayloadNONCE_1,
    enums.Payload.NOTIFY_1: PayloadNOTIFY_1,
    enums.Payload.DELETE_1: PayloadDELETE_1,
    enums.Payload.VENDOR_1: PayloadVENDOR_1,
    enums.Payload.CP_1: PayloadCP_1,
    enums.Payload.NATD_1: PayloadNATD_1,
    enums.Payload.SA: PayloadSA,
    enums.Payload.KE: PayloadKE,
    enums.Payload.IDi: PayloadIDi,
    enums.Payload.IDr: PayloadIDr,
    enums.Payload.AUTH: PayloadAUTH,
    enums.Payload.NONCE: PayloadNONCE,
    enums.Payload.NOTIFY: PayloadNOTIFY,
    enums.Payload.DELETE: PayloadDELETE,
    enums.Payload.VENDOR: PayloadVENDOR,
    enums.Payload.TSi: PayloadTSi,
    enums.Payload.TSr: PayloadTSr,
    enums.Payload.SK: PayloadSK,
    enums.Payload.CP: PayloadCP,
    enums.Payload.EAP: PayloadEAP,
}

class Message:
    def __init__(self, spi_i, spi_r, version, exchange, flag, message_id, payloads=None, *, first_payload=None):
        self.spi_i = spi_i
        self.spi_r = spi_r
        self.version = version
        self.exchange = enums.Exchange(exchange)
        self.flag = enums.MsgFlag(flag)
        self.message_id = message_id
        self.first_payload = first_payload
        self.payloads = [] if payloads is None else payloads
    @classmethod
    def parse(cls, stream):
        header = struct.unpack('>8s8s4B2L', stream.read(28))
        return Message(header[0], header[1], header[3], header[4], header[5], header[6], first_payload=header[2])
    def parse_payloads(self, stream, *, crypto=None):
        if self.flag & enums.MsgFlag.Encryption:
            stream = io.BytesIO(crypto.decrypt_1(stream.read(), self.message_id))
        next_payload = self.first_payload
        while next_payload:
            payload_id = next_payload
            next_payload, critical, length = struct.unpack('>BBH', stream.read(4))
            critical = bool(critical >> 7)
            payload = PayloadClass.get(payload_id, Payload).parse(payload_id, critical, stream, length-4)
            if payload_id == enums.Payload.SK:
                if crypto is not None:
                    crypto.verify_checksum(stream.getvalue())
                    decrypted = crypto.decrypt(payload.ciphertext)
                    stream = io.BytesIO(decrypted)
                    continue
                payload.next_payload = next_payload
                next_payload = enums.Payload.NONE
            self.payloads.append(payload)
    @classmethod
    def encode_payloads(cls, payloads):
        data = bytearray()
        for idx, payload in enumerate(payloads):
            payload_data = payload.to_bytes()
            if idx < len(payloads) - 1:
                next_payload = payloads[idx+1].type
            else:
                next_payload = enums.Payload.NONE
            data.extend(struct.pack('>BBH', next_payload, 0x80 if payload.critical else 0x00, len(payload_data) + 4))
            data.extend(payload_data)
        return data
    def to_bytes(self, *, crypto=None):
        first_payload = self.payloads[0].type if self.payloads else enums.Payload.NONE
        data = self.encode_payloads(self.payloads)
        if crypto and self.version == 0x10:
            data = bytearray(crypto.encrypt_1(data, self.message_id))
            self.flag |= enums.MsgFlag.Encryption
        elif crypto and self.version == 0x20:
            payload_sk = PayloadSK(crypto.encrypt(data))
            payload_data = payload_sk.to_bytes()
            data = bytearray(struct.pack('>BxH', first_payload, len(payload_data) + 4) + payload_data)
            first_payload = enums.Payload.SK
        data[0:0] = struct.pack(
            '>8s8s4B2L', self.spi_i, self.spi_r, first_payload,
            self.version, self.exchange, self.flag,
            self.message_id, 28+len(data))
        if crypto and self.version == 0x20:
            crypto.add_checksum(data)
        return data
    def __repr__(self):
        return f'{self.exchange.name}(spi_i={self.spi_i.hex()}, spi_r={self.spi_r.hex()}, version={self.version>>4}.{self.version&0xF}, flag={self.flag!s}, message_id={self.message_id}, ' + \
                (', '.join(repr(i) for i in self.payloads) or 'NONE') + ')'
    def get_payload(self, payload_type):
        return next((x for x in self.payloads if x.type == payload_type), None)
    def get_payload_notify(self, notify_id):
        return next((x for x in self.payloads if x.type == enums.Payload.NOTIFY and x.notify == notify_id), None)



from setuptools import setup
import os, re

def read(*names, **kwargs):
    with open(os.path.join(os.path.dirname(__file__), *names), encoding='utf8') as fp:
        return fp.read()

def find_value(name):
    data_file = read('pvpn', '__doc__.py')
    data_match = re.search(r"^__%s__ += ['\"]([^'\"]*)['\"]" % name, data_file, re.M)
    if data_match:
        return data_match.group(1)
    raise RuntimeError(f"Unable to find '{name}' string.")

setup(
    name                = find_value('title'),
    version             = find_value('version'),
    description         = find_value('description'),
    long_description    = read('README.rst'),
    url                 = find_value('url'),
    author              = find_value('author'),
    author_email        = find_value('email'),
    license             = find_value('license'),
    python_requires     = '>=3.6',
    keywords            = find_value('keywords'),
    packages            = ['pvpn'],
    classifiers         = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    install_requires    = [
        'pycryptodome >= 3.7.2',
        'pproxy >= 2.0.0',
    ],
    entry_points        = {
        'console_scripts': [
            'pvpn = pvpn.server:main',
        ],
    },
)
